"""
======================COPYRIGHT/LICENSE START==========================

generalIO.py: General I/O information and code for nmrStar files.

Copyright (C) 2005-2010 Wim Vranken (European Bioinformatics Institute)

=======================================================================

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

A copy of this license can be found in ../../../../license/LGPL.license

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA


======================COPYRIGHT/LICENSE END============================

for further information, please contact :

- CCPN website (http://www.ccpn.ac.uk/)
- PDBe website (http://www.ebi.ac.uk/pdbe/)

- contact Wim Vranken (wim@ebi.ac.uk)
=======================================================================

If you are using this software for academic purposes, we suggest
quoting the following references:

===========================REFERENCE START=============================
R. Fogh, J. Ionides, E. Ulrich, W. Boucher, W. Vranken, J.P. Linge, M.
Habeck, W. Rieping, T.N. Bhat, J. Westbrook, K. Henrick, G. Gilliland,
H. Berman, J. Thornton, M. Nilges, J. Markley and E. Laue (2002). The
CCPN project: An interim report on a data model for the NMR community
(Progress report). Nature Struct. Biol. 9, 416-418.

Wim F. Vranken, Wayne Boucher, Tim J. Stevens, Rasmus
H. Fogh, Anne Pajon, Miguel Llinas, Eldon L. Ulrich, John L. Markley, John
Ionides and Ernest D. Laue (2005). The CCPN Data Model for NMR Spectroscopy:
Development of a Software Pipeline. Proteins 59, 687 - 696.

===========================REFERENCE END===============================
"""

import os, sys, string

from b2bTools.general.ccpn.universal.Io import getTopDirectory

from b2bTools.general.ccpn.format.general.formatIO import FormatFile

from b2bTools.general.ccpn.format.nmrStar.util import returnStarInt
from b2bTools.general.ccpn.format.nmrStar.util import returnStarFloat
from b2bTools.general.ccpn.format.nmrStar.util import returnStarString
from b2bTools.general.ccpn.format.nmrStar.util import getNmrStarValue

# Import Jurgen Doreleijers' Star reader/writer stuff
import b2bTools.general.bmrb.File as nmrStar

from b2bTools.general.ccpn.format.general.Constants import defaultMolCode

"""

Notes for rewriting this script for more generic nmrStar treatment:


IF ANY OF BELOW INCLUDES ACTUAL CCPNMR STUFF THEN DO NOT PUT IT HERE!!

- separate reference data classes from 'real' ones, put link in 'real' ones to
  reference data

- make loops to set reoccuring tags (e.g. atom_name_1, ...)

- for reference data make table and tag classes!!! Use these (also keep ..Names list for
  quick string checking)

- add 'attribute' names to sfDict (or tag class)

- Set attribute for ref SF class: isCcpMapped (if false, store as application data linked to Entry!)

- Set attribute for ref SF tables: isCcpMapped (if false, store as application data to
  whatever parent object is given during interpretation. Could in principle set this as well...)

- Set attribute for ref SF tags? Is more difficult to set... also give parent object as for tables?
  could then check for appldata on that level... .



- Kinds of tags:

   - not mapped, use if exists as application data

   - mapped with direct mapping
   - mapped with function mapping
   - mapped with no direct or function mapping (comes from function at other tag, e.g. chainlabel from atom)
   - each/some of the 'mapped' could also exist as application data? Need flag to know what to use...

- example: Label_atom_ID has function 'getAtomLabels' which ALSO sets the
          Label_entity_assembly_ID, Label_entity_ID, Label_comp_index_ID and Label_comp_ID
          (these tags then have no function or direct mapping but are labelled as mapped)



Could do setup of saveframes, tables, tags as reference classes (also keep data in there?),
then have functions linked to sets of tags to handle multiple ones (e.g. entity_assembly, entity, ...)

Organize functions for input as datamodel classes, output as saveframe/... ?


"""

#####################
# Class definitions #
#####################

class NmrStarGenericFile(FormatFile):

  def setGeneric(self):

    self.format = 'nmrStar'
    self.defaultMolCode = defaultMolCode
    self.tagSep = '.'

    self.measureExpts = []
    self.measureSofts = []

    self.structSofts = []
    self.structMeths = []

    if not self.version:
      self.version = '2.1.1' # is default

    # This is a hack so can modify an input file before reading. Useful for chemical shift files
    # where often only the loop_ is reported by programs (e.g. CARA).
    self.text = ""

  def checkVersion(self):

    if self.version not in ['2.1.1','3.0','3.1']:

      print ("  Cannot parse nmrStar distance constraints for version %s" % self.version)
      return False

    return True

  def setMeasureExperiments(self,tableName):

    if tableName in self.saveFrame.tables:
      exptTableTags = self.saveFrame.tables[tableName].tags
      numExpts = len(exptTableTags['Experiment_ID'])

      for i in range(0,numExpts):

        self.measureExpts.append(NmrStarMeasureExpt(self))
        self.measureExpts[-1].setData(exptTableTags,i)

  def setMeasureSoftwares(self,tableName):

    if tableName in self.saveFrame.tables:
      softwareTableTags = self.saveFrame.tables[tableName].tags
      numSoft = len(softwareTableTags['Software_ID'])

      for i in range(0,numSoft):

        self.measureSofts.append(NmrStarMeasureSoft(self))
        self.measureSofts[-1].setData(softwareTableTags,i)

  def setStructSoftwares(self,tableName):

    if tableName in self.saveFrame.tables:
      softwareTableTags = self.saveFrame.tables[tableName].tags
      numSoft = len(softwareTableTags['Software_ID'])

      for i in range(0,numSoft):

        self.structSofts.append(NmrStarStructSoft(self))
        self.structSofts[-1].setData(softwareTableTags,i)

  def setStructMethods(self,tableName):

    if tableName in self.saveFrame.tables:
      methodTableTags = self.saveFrame.tables[tableName].tags
      numMeth = len(methodTableTags['Entry_ID'])

      for i in range(0,numMeth):

        self.structMeths.append(NmrStarStructMeth(self))
        self.structMeths[-1].setData(methodTableTags,i)

class NmrStarMeasureExpt:

  def __init__(self,parent):

    self.parent = parent

  def setData(self,exptTableTags,i):

    assignList = [['exptId',      'Experiment_ID', None],
                  ['exptName',    'Experiment_name', None],
                  ['sampleId',    'Sample_ID', None],
                  ['sampleState', 'Sample_state', None]
                  ]

    for (attrName,tagName,default) in assignList:

      if tagName in exptTableTags:
        if exptTableTags[tagName][i] != None:
          setattr(self,attrName,exptTableTags[tagName][i])
        else:
          setattr(self,attrName,default)

class NmrStarMeasureSoft:

  def __init__(self,parent):

    self.parent = parent

  def setData(self,softTableTags,i):

    assignList = [['softId',      'Software_ID', None],
                  ['softName',    'Software_label', None],
                  ['methodId',    'Method_ID', None],
                  ['methodLabel', 'Method_label', None]
                  ]

    for (attrName,tagName,default) in assignList:

      if tagName in softTableTags:
        if softTableTags[tagName][i] != None:
          setattr(self,attrName,softTableTags[tagName][i])
        else:
          setattr(self,attrName,default)

class NmrStarStructSoft:

  def __init__(self,parent):

    self.parent = parent

  def setData(self,softTableTags,i):

    assignList = [['softId',      'Software_ID', None],
                  ['softName',    'Software_label', None],
                  ['methodId',    'Method_ID', None],
                  ['methodLabel', 'Method_label', None]
                  ]

    for (attrName,tagName,default) in assignList:

      if tagName in softTableTags:
        if softTableTags[tagName][i] != None:
          setattr(self,attrName,softTableTags[tagName][i])
        else:
          setattr(self,attrName,default)

class NmrStarStructMeth:

  def __init__(self,parent):

    self.parent = parent

  def setData(self,methTableTags,i):

    assignList = [['methodLabel',   'Refine_method', None],
                  ['methodDetails', 'Refine_details', None]
                  ]

    for (attrName,tagName,default) in assignList:

      if tagName in methTableTags:
        if methTableTags[tagName][i] != None:
          setattr(self,attrName,methTableTags[tagName][i])
        else:
          setattr(self,attrName,default)

class NmrStarConstraintFile(NmrStarGenericFile):

  def parseCommentsLoop(self):

    self.comments = []

    #
    # Parse the comments loop (not really necessary but good example)
    #

    tableTagNames = {}
    tableName = None

    if self.version == '2.1.1':

      tableName = '_Constraint_comment_ID'
      tableTagNames['commentID'] = tableName

    elif self.version in ('3.0','3.1'):

      commentTableText = 'constraint_comment_org'

      tableName = self.saveFrame.findTableByEndText(commentTableText)
      tableTagNames['commentID'] = 'ID'

    if tableName and tableName in self.saveFrame.tables:

      constraintTableTags = self.saveFrame.tables[tableName].tags

      constraintTableTagKeys = constraintTableTags.keys()

      self.commentTags = constraintTableTagKeys

      for i in range(0,len(constraintTableTags[tableTagNames['commentID']])):

        commentTags = []

        for tagKey in constraintTableTagKeys:
          commentTags.append(constraintTableTags[tagKey][i])

        self.comments.append(commentTags)

class NmrStarFile(NmrStarGenericFile):

  def initialize(self):

    self.components = ['all']

  def setComponentList(self):

    pass

  def setComponents(self):

    self.setComponentList()

    # This should become list of stuff
    self.sfs = {}

    #
    # nmrStar DEFINITIONS (to get right value type)
    #
    #  Could be extended to ccpn directly... check this later
    #

    self.sfDict = {}

    for component in self.components:

      self.setSfDict(component)

  def setSfDict(self,component):

    #
    # Keep track of component order
    #

    self.componentList = []

    if self.version == '2.1.1':

      value = 0

      #
      # This loop occurs multiple times: defined here once
      #

      constraintCommentLoop = [

            '_Constraint_comment_ID', [

                ['_Constraint_comment_ID',           lambda x = value: returnStarInt(x)],
                ['_Constraint_comment',              lambda x = value: returnStarString(x)],
                ['_Constraint_comment_begin_line',   lambda x = value: returnStarInt(x)],
                ['_Constraint_comment_begin_column', lambda x = value: returnStarInt(x)],
                ['_Constraint_comment_end_line',     lambda x = value: returnStarInt(x)],
                ['_Constraint_comment_end_column',   lambda x = value: returnStarInt(x)]
              ]
            ]

      #
      # Entry information
      #

      sfName = 'entry_information'

      if component == sfName or component == 'all':

        self.componentList.append(sfName)
        self.sfDict[sfName] = SaveFrame("",sfName)

        self.sfDict[sfName].tags = [

            ['_Entry_title',          lambda x = value: x],
            #['_Version_type',         lambda x = value: returnStarString(x)],
            ['_NMR_STAR_version',     lambda x = value: returnStarString(x)],
            ['_Submission_date',      lambda x = value: returnStarString(x)],
            ['_Accession_date',       lambda x = value: returnStarString(x)],
            ['_Experimental_method',  lambda x = value: returnStarString(x)]

          ]

        # Warning: key here is tagName by which table is identified!
        self.sfDict[sfName].tables = [

            ['_Author_ordinal', [

                ['_Author_ordinal',           lambda x = value: returnStarInt(x)],
                ['_Author_family_name',       lambda x = value: returnStarString(x)],
                ['_Author_given_name',        lambda x = value: returnStarString(x)],
                ['_Author_middle_initials',   lambda x = value: returnStarString(x)],
                ['_Author_family_title',      lambda x = value: returnStarString(x)]
              ]
            ],

            ['_Related_BMRB_accession_number', [

                ['_Related_BMRB_accession_number',   lambda x = value: returnStarInt(x)],
                ['_Relationship',                    lambda x = value: returnStarString(x)],
              ]
            ]

          ]


      #
      # Entry citation
      #

      sfName = 'entry_citation'

      if component == sfName or component == 'all':

        self.componentList.append(sfName)
        self.sfDict[sfName] = SaveFrame("",sfName)

        self.sfDict[sfName].tags = [

            ['_Citation_title',    lambda x = value: x],
            ['_Citation_status',   lambda x = value: returnStarString(x)],
            ['_Citation_type',     lambda x = value: returnStarString(x)]

          ]

        # Warning: key here is tagName by which table is identified!
        self.sfDict[sfName].tables = [
            # Can add author info here if required...
            ['_Keyword', [
                ['_Keyword',  lambda x = value: returnStarString(x)]
              ]
            ]
          ]

      #
      # Constraint file definitions Jurgen MR format
      #

      sfName = 'MR_file_characteristics'

      if component == sfName or component == 'all':

        self.componentList.append(sfName)
        self.sfDict[sfName] = SaveFrame("",sfName)

        self.sfDict[sfName].tags = [

            ['_Saveframe_category',             lambda x = value: x],
            ['_BMRB_dev_PDB_id',                lambda x = value: returnStarString(x)]
          ]


        # Warning: key here is tagName by which table is identified!
        self.sfDict[sfName].tables = [

            ['_BMRB_dev_Position_block_in_MR_file', [

                ['_BMRB_dev_Position_block_in_MR_file',  lambda x = value: returnStarInt(x)],
                ['_BMRB_dev_Program',                    lambda x = value: returnStarString(x)],
                ['_BMRB_dev_Type',                       lambda x = value: returnStarString(x)],
                ['_BMRB_dev_Subtype',                    lambda x = value: returnStarString(x)],
                ['_BMRB_dev_Format',                     lambda x = value: returnStarString(x)]
              ]
            ]
          ]

      #
      # Comments Jurgen MR format
      #

      sfName = 'MR_file_comments'

      if component == sfName or component == 'all':

        self.componentList.append(sfName)
        self.sfDict[sfName] = SaveFrame("",sfName)

        self.sfDict[sfName].tags = [

            ['_Saveframe_category',                 lambda x = value: returnStarString(x)],
            ['_BMRB_dev_Position_block_in_MR_file', lambda x = value: returnStarInt(x)],
            ['_BMRB_dev_Comment',                   lambda x = value: returnStarString(x)],

          ]

      #
      # Molecular system info
      #
      # Warning: not complete! Quick hack to get info out of 2.1.1 files...
      #

      sfName = 'molecular_system'

      if component == sfName or component == 'all':

        self.componentList.append(sfName)
        self.sfDict[sfName] = SaveFrame("",sfName)

        self.sfDict[sfName].tags = [

            ['_Mol_system_name',                  lambda x = value: x],
            ['_Abbreviation_common',              lambda x = value: x],
            ['_Enzyme_commission_number',         lambda x = value: x],
            ['_System_physical_state',            lambda x = value: x],
            ['_System_oligomer_state',            lambda x = value: x],
            ['_System_paramagnetic',              lambda x = value: x],
            ['_System_thiol_state',               lambda x = value: x]

          ]


        # Warning: key here is tagName by which table is identified!
        self.sfDict[sfName].tables = [

            ['_Mol_system_component_name', [

                ['_Mol_system_component_name',   lambda x = value: x],
                ['_Mol_label',                   lambda x = value: x]

              ]
            ],

            ['_Biological_function', [

                ['_Biological_function',                          lambda x = value: x]

              ]
            ]
          ]

      #
      # Sequence for a monomeric polymer
      #

      sfName = 'monomeric_polymer'

      if component == sfName or component == 'all':

        self.componentList.append(sfName)
        self.sfDict[sfName] = SaveFrame("",sfName)

        self.sfDict[sfName].tags = [

            ['_Mol_type',                       lambda x = value: x],
            ['_Mol_polymer_class',              lambda x = value: x],
            ['_Name_common',                    lambda x = value: x],
            ['_Name_variant',                   lambda x = value: x],
            ['_Name_common',                    lambda x = value: x],
            ['_Abbreviation_common',            lambda x = value: x],
            ['_Molecular_mass',                 lambda x = value: returnStarFloat(x)],
            ['_Mol_thiol_state',                lambda x = value: x],
            ['_Residue_count',                  lambda x = value: returnStarInt(x)],
            ['_Mol_residue_sequence',           lambda x = value: x]
          ]


        # Warning: key here is tagName by which table is identified!
        self.sfDict[sfName].tables = [

            ['_Residue_seq_code', [

                ['_Residue_seq_code',           lambda x = value: returnStarInt(x)],
                ['_Residue_author_seq_code',    lambda x = value: returnStarInt(x)],
                ['_Residue_label',              lambda x = value: x]
              ]
            ],

            ['_Database_name', [

                ['_Database_name',                          lambda x = value: x],
                ['_Database_accession_code',                lambda x = value: x],
                ['_Database_entry_mol_name',                lambda x = value: x],
                ['_Sequence_query_to_submitted_percentage', lambda x = value: returnStarFloat(x)],
                ['_Sequence_subject_length',                lambda x = value: returnStarInt(x)],
                ['_Sequence_identity',                      lambda x = value: returnStarFloat(x)],
                ['_Sequence_positive',                      lambda x = value: returnStarFloat(x)],
                ['_Sequence_homology_expectation_value',    lambda x = value: returnStarFloat(x)]

              ]
            ]
          ]

      #
      # Ligands
      #

      sfName = 'ligand'

      if component == sfName or component == 'all':

        self.componentList.append(sfName)
        self.sfDict[sfName] = SaveFrame("",sfName)

        self.sfDict[sfName].tags = [

            ['_Mol_type',                       lambda x = value: x],
            ['_Name_common',                    lambda x = value: x],
            ['_Abbreviation_common',            lambda x = value: x],
            ['_Name_IUPAC',                     lambda x = value: x],
            ['_BMRB_code',                      lambda x = value: x],
            ['_PDB_code',                       lambda x = value: x],
            ['_Mol_empirical_formula',          lambda x = value: x],
            ['_Mol_paramagnetic',               lambda x = value: x],
            ['_Molecular_mass',                 lambda x = value: returnStarFloat(x)],
            ['_Mol_charge',                     lambda x = value: returnStarInt(x)],
            ['_Details',                        lambda x = value: x]
          ]

      #
      # Assigned chemical shifts
      #

      sfName = 'assigned_chemical_shifts'

      if component == sfName or component == 'all':

        self.componentList.append(sfName)
        self.sfDict[sfName] = SaveFrame("",sfName)

        self.sfDict[sfName].tags = [

            ['_Saveframe_category',             lambda x = value: x],
            ['_Mol_system_component_name',      lambda x = value: x],
            ['_Sample_conditions_label',        lambda x = value: x],
            ['_Chem_shift_reference_set_label', lambda x = value: x]
          ]


        # Warning: key here is tagName by which table is identified!
        self.sfDict[sfName].tables = [

            ['_Sample_label', [

                ['_Sample_label',      lambda x = value: x]

              ]
            ],

            ['_Atom_shift_assign_ID', [

                ['_Atom_shift_assign_ID',      lambda x = value: returnStarInt(x)],
                ['_Residue_author_seq_code',   lambda x = value: returnStarInt(x)],
                ['_Residue_seq_code',          lambda x = value: returnStarInt(x)],
                ['_Residue_label',             lambda x = value: x],
                ['_Atom_name',                 lambda x = value: x],
                ['_Atom_type',                 lambda x = value: x],
                ['_Chem_shift_value',          lambda x = value: returnStarFloat(x)],
                ['_Chem_shift_value_error',    lambda x = value: returnStarFloat(x)],
                ['_Chem_shift_ambiguity_code', lambda x = value: returnStarInt(x)],
                ['_Chem_shift_ambiguity_type', lambda x = value: returnStarInt(x)]    # This also sometimes occurs...
              ]
            ]
          ]

      #
      # J couplings
      #

      sfName = 'coupling_constants'

      if component == sfName or component == 'all':

        self.componentList.append(sfName)
        self.sfDict[sfName] = SaveFrame("",sfName)

        self.sfDict[sfName].tags = [

            ['_Saveframe_category',             lambda x = value: x],
            ['_Mol_system_component_name',      lambda x = value: x],
            ['_Sample_conditions_label',        lambda x = value: x],
          ]


        # Warning: key here is tagName by which table is identified!
        self.sfDict[sfName].tables = [

            ['_Sample_label', [

                ['_Sample_label',      lambda x = value: x]

              ]
            ],

            ['_Coupling_constant_code', [

                ['_Coupling_constant_code',    lambda x = value: x],
                ['_Atom_one_residue_seq_code', lambda x = value: returnStarInt(x)],
                ['_Atom_one_residue_label',    lambda x = value: x],
                ['_Atom_one_name',             lambda x = value: x],
                ['_Atom_two_residue_seq_code', lambda x = value: returnStarInt(x)],
                ['_Atom_two_residue_label',    lambda x = value: x],
                ['_Atom_two_name',             lambda x = value: x],
                ['_Coupling_constant_value',       lambda x = value: returnStarFloat(x)],
                ['_Coupling_constant_value_error', lambda x = value: returnStarFloat(x)],
                ['_Coupling_constant_min_value',   lambda x = value: returnStarFloat(x)],
                ['_Coupling_constant_max_value',   lambda x = value: returnStarFloat(x)],
              ]
            ]
          ]


      #
      # Order parameters
      #

      sfName = 'S2_parameters'

      if component == sfName or component == 'all':

        self.componentList.append(sfName)
        self.sfDict[sfName] = SaveFrame("",sfName)

        self.sfDict[sfName].tags = [

            ['_Saveframe_category',             lambda x = value: x],
            ['_Mol_system_component_name',      lambda x = value: x],
            ['_Sample_conditions_label',        lambda x = value: x],
          ]


        # Warning: key here is tagName by which table is identified!
        self.sfDict[sfName].tables = [

            ['_Sample_label', [

                ['_Sample_label',      lambda x = value: x]

              ]
            ],

            ['_Software_label', [

                ['_Software_label',      lambda x = value: x]

              ]
            ],

            ['_Residue_seq_code', [

                ['_Residue_seq_code', lambda x = value: returnStarInt(x)],
                ['_Residue_label',    lambda x = value: x],
                ['_Atom_name',        lambda x = value: x],
                ['_Model_fit',        lambda x = value: x],
                ['_S2_value',              lambda x = value: returnStarFloat(x)],
                ['_S2_value_fit_error',    lambda x = value: returnStarFloat(x)],
                ['_Tau_e_value',           lambda x = value: returnStarFloat(x)],
                ['_Tau_e_value_fit_error', lambda x = value: returnStarFloat(x)],
                ['_S2f_value',             lambda x = value: returnStarFloat(x)],
                ['_S2f_value_fit_error',   lambda x = value: returnStarFloat(x)],
                ['_S2s_value',             lambda x = value: returnStarFloat(x)],
                ['_S2s_value_fit_error',   lambda x = value: returnStarFloat(x)],
                ['_Tau_s_value',           lambda x = value: returnStarFloat(x)],
                ['_Tau_s_value_fit_error', lambda x = value: returnStarFloat(x)],
                ['_S2H_value',             lambda x = value: returnStarFloat(x)],
                ['_S2H_value_fit_error',   lambda x = value: returnStarFloat(x)],
                ['_S2N_value',             lambda x = value: returnStarFloat(x)],
                ['_S2N_value_fit_error',   lambda x = value: returnStarFloat(x)],
              ]
            ]
          ]

      #
      # Sample conditions
      #

      sfName = 'sample_conditions'

      if component == sfName or component == 'all':

        self.componentList.append(sfName)
        self.sfDict[sfName] = SaveFrame("",sfName)

        self.sfDict[sfName].tags = [
            ['_Saveframe_category',             lambda x = value: x]
          ]


        # Warning: key here is tagName by which table is identified!
        self.sfDict[sfName].tables = [

            ['_Variable_type', [

                ['_Variable_type',        lambda x = value: x],
                ['_Variable_value',       lambda x = value: returnStarFloat(x)],
                ['_Variable_value_error', lambda x = value: returnStarFloat(x)],
                ['_Variable_value_units', lambda x = value: x]

              ]
            ]
          ]

      #
      # Sample
      #

      sfName = 'sample'

      if component == sfName or component == 'all':

        self.componentList.append(sfName)
        self.sfDict[sfName] = SaveFrame("",sfName)

        self.sfDict[sfName].tags = [
            ['_Saveframe_category',             lambda x = value: x],
            ['_Sample_type',                    lambda x = value: x]
          ]


        # Warning: key here is tagName by which table is identified!
        self.sfDict[sfName].tables = [

            ['_Mol_label', [

                ['_Mol_label',                 lambda x = value: x],
                ['_Concentration_value',       lambda x = value: returnStarFloat(x)],
                ['_Concentration_value_units', lambda x = value: x],
                ['_Concentration_min_value',   lambda x = value: returnStarFloat(x)],
                ['_Concentration_max_value',   lambda x = value: returnStarFloat(x)],
                ['_Isotopic_labeling',         lambda x = value: x]

              ]
            ]
          ]

      #
      # Chemical shift reference
      #

      sfName = 'chemical_shift_reference'

      if component == sfName or component == 'all':

        self.componentList.append(sfName)
        self.sfDict[sfName] = SaveFrame("",sfName)

        self.sfDict[sfName].tags = [

            ['_Saveframe_category',  lambda x = value: x],
            ['_Details',             lambda x = value: x],
         ]

        # Warning: key here is tagName by which table is identified!
        self.sfDict[sfName].tables = [

            ['_Mol_common_name', [

                ['_Mol_common_name',                    lambda x = value: x],
                ['_Atom_type',                          lambda x = value: x],
                ['_Atom_isotope_number',                lambda x = value: returnStarInt(x)],
                ['_Atom_group',                         lambda x = value: x],
                ['_Chem_shift_units',                   lambda x = value: x],
                ['_Chem_shift_value',                   lambda x = value: returnStarFloat(x)],
                ['_Reference_method',                   lambda x = value: x],
                ['_Reference_type',                     lambda x = value: x],
                ['_External_reference_sample_geometry', lambda x = value: x],
                ['_External_reference_location',        lambda x = value: x],
                ['_External_reference_axis',            lambda x = value: x],
                ['_Indirect_shift_ratio',               lambda x = value: returnStarFloat(x)]
              ]
            ]
          ]

      #
      # Distance constraints (Jurgen format)
      #

      sfName = 'distance_constraints'

      if component == sfName or component == 'all':

        self.componentList.append(sfName)
        self.sfDict[sfName] = SaveFrame("",sfName)

        self.sfDict[sfName].tags = [

            ['_Saveframe_category',             lambda x = value: x],
            ['_BMRB_dev_Position_block_in_MR_file', lambda x = value: returnStarInt(x)]
          ]


        # Warning: key here is tagName by which table is identified!
        self.sfDict[sfName].tables = [

            ['_Constraint_ID', [

                ['_Constraint_ID',                   lambda x = value: returnStarInt(x)],
                ['_Constraint_tree_node_ID',         lambda x = value: returnStarInt(x)],
                ['_Constraint_tree_logic_operation', lambda x = value: returnStarString(x,strip=1)],
                ['_Constraint_tree_down_node_ID',    lambda x = value: returnStarInt(x)],
                ['_Constraint_tree_right_node_ID',   lambda x = value: returnStarInt(x)]
              ]
            ],

            ['_Constraint_tree_node_member_constraint_ID', [

                ['_Constraint_tree_node_member_constraint_ID', lambda x = value: returnStarInt(x)],
                ['_Constraint_tree_node_member_node_ID',       lambda x = value: returnStarInt(x)],
                ['_Constraint_tree_node_member_ID',            lambda x = value: returnStarInt(x)],
                ['_Contribution_fractional_value',             lambda x = value: returnStarString(x)],
                ['_Mol_system_component_code',                 lambda x = value: returnStarString(x,strip=1)],
                ['_Residue_seq_code',                          lambda x = value: returnStarInt(x)],
                ['_Atom_name',                                 lambda x = value: returnStarString(x,strip=1)]
              ]
            ],

            ['_Distance_constraint_ID', [

                ['_Distance_constraint_ID',           lambda x = value: returnStarInt(x)],
                ['_Distance_constraint_tree_node_ID', lambda x = value: returnStarInt(x)],
                ['_Distance_value',                   lambda x = value: returnStarFloat(x)],
                ['_Distance_lower_bound_value',       lambda x = value: returnStarFloat(x)],
                ['_Distance_upper_bound_value',       lambda x = value: returnStarFloat(x)]
              ]
            ],

            constraintCommentLoop
          ]

      #
      # Dihedral constraints (Jurgen format or old?)
      #

      sfName = 'torsion_angle_constraints'

      if component == sfName or component == 'all':

        self.componentList.append(sfName)
        self.sfDict[sfName] = SaveFrame("",sfName)

        self.sfDict[sfName].tags = [

            ['_Saveframe_category',                 lambda x = value: x],
            ['_BMRB_dev_Position_block_in_MR_file', lambda x = value: returnStarInt(x)],
            ['_BMRB_dev_Program',                   lambda x = value: x],
            ['_BMRB_dev_Type',                      lambda x = value: x],
            ['_BMRB_dev_Subtype',                   lambda x = value: x],
            ['_BMRB_dev_Format',                    lambda x = value: x]
          ]


        # Warning: key here is tagName by which table is identified!
        self.sfDict[sfName].tables = [

            ['_Constraint_ID', [

                ['_Constraint_ID',                    lambda x = value: returnStarInt(x)],
                ['_Angle_name',                       lambda x = value: returnStarString(x,strip=1)],
                ['_Atom_one_mol_system_component_ID', lambda x = value: returnStarString(x,strip=1)],
                ['_Atom_one_residue_seq_code',        lambda x = value: returnStarInt(x)],
                ['_Atom_one_residue_label',           lambda x = value: returnStarString(x,strip=1)],
                ['_Atom_one_atom_name',             lambda x = value: returnStarString(x,strip=1)],
                ['_Atom_two_mol_system_component_ID', lambda x = value: returnStarString(x,strip=1)],
                ['_Atom_two_residue_seq_code',        lambda x = value: returnStarInt(x)],
                ['_Atom_two_residue_label',           lambda x = value: returnStarString(x,strip=1)],
                ['_Atom_two_atom_name',             lambda x = value: returnStarString(x,strip=1)],
                ['_Atom_three_mol_system_component_ID',lambda x = value: returnStarString(x,strip=1)],
                ['_Atom_three_residue_seq_code',      lambda x = value: returnStarInt(x)],
                ['_Atom_three_residue_label',         lambda x = value: returnStarString(x,strip=1)],
                ['_Atom_three_atom_name',             lambda x = value: returnStarString(x,strip=1)],
                ['_Atom_four_mol_system_component_ID',lambda x = value: returnStarString(x,strip=1)],
                ['_Atom_four_residue_seq_code',       lambda x = value: returnStarInt(x)],
                ['_Atom_four_residue_label',          lambda x = value: returnStarString(x,strip=1)],
                ['_Atom_four_atom_name',            lambda x = value: returnStarString(x,strip=1)],
                ['_Angle_upper_bound_value',          lambda x = value: returnStarFloat(x)],
                ['_Angle_lower_bound_value',          lambda x = value: returnStarFloat(x)],
                ['_Force_constant_value',             lambda x = value: returnStarFloat(x)],
                ['_Potential_function_exponent',      lambda x = value: returnStarInt(x)]

              ]
            ],

            constraintCommentLoop
          ]

      #
      # Residual dipolar coupling values
      #

      sfName = 'residual_dipolar_couplings'

      if component == sfName or component == 'all':

        self.componentList.append(sfName)
        self.sfDict[sfName] = SaveFrame("",sfName)

        self.sfDict[sfName].tags = [

            ['_Saveframe_category',             lambda x = value: x],
            ['_Sample_conditions_label',        lambda x = value: x],
            ['_Sample_conditions',              lambda x = value: x],   # Warning; this is not correct but often occurs, so hacking it in...
            ['_Spectrometer_frequency_1H',      lambda x = value: returnStarFloat(x)],
          ]


        # Warning: key here is tagName by which table is identified!
        self.sfDict[sfName].tables = [

            ['_Sample_label', [

                ['_Sample_label',      lambda x = value: x]

              ]
            ],

            ['_Residual_dipolar_coupling_ID', [

                ['_Residual_dipolar_coupling_ID',     lambda x = value: returnStarInt(x)],
                ['_Residual_dipolar_coupling_code',   lambda x = value: returnStarString(x)],
                ['_Atom_one_mol_system_component_name',lambda x = value: returnStarString(x,strip=1)],
                ['_Atom_one_residue_seq_code',        lambda x = value: returnStarInt(x)],
                ['_Atom_one_residue_label',           lambda x = value: returnStarString(x,strip=1)],
                ['_Atom_one_atom_name',               lambda x = value: returnStarString(x,strip=1)],
                ['_Atom_two_mol_system_component_name',lambda x = value: returnStarString(x,strip=1)],
                ['_Atom_two_residue_seq_code',        lambda x = value: returnStarInt(x)],
                ['_Atom_two_residue_label',           lambda x = value: returnStarString(x,strip=1)],
                ['_Atom_two_atom_name',               lambda x = value: returnStarString(x,strip=1)],
                ['_Residual_dipolar_coupling_value',  lambda x = value: returnStarFloat(x)],
                ['_Residual_dipolar_coupling_error',  lambda x = value: returnStarFloat(x)],
                ['_Residual_dipolar_coupling_value_error',  lambda x = value: returnStarFloat(x)],
                ['_Residual_dipolar_coupling_lower_bound_value',lambda x = value: returnStarFloat(x)],
                ['_Residual_dipolar_coupling_upper_bound_value',lambda x = value: returnStarFloat(x)]

              ]
            ],

            constraintCommentLoop
          ]

    ###################################
    #                                 #
    # START OF VERSION 3.0 STUFF!!    #
    #                                 #
    ###################################

    elif self.version == '3.0':

      value = 0

      #
      # TODO: set this up from v3.0 dictionary!!
      #
      # Also: automatically reocurring tags in loops
      # should be handled from sf level (and not defined in dict,
      # or defined by a function when writing!!!)
      #

      #
      # Constraint file definitions version 3.0
      #

      sfName = 'file_characteristics'

      if component == sfName or component == 'all':

        self.componentList.append(sfName)
        self.sfDict[sfName] = SaveFrame("",sfName)

        self.sfDict[sfName].tags = [

            ['Sf_category',          lambda x = value: x],
            ['ID',                   lambda x = value: returnStarInt(x)],
            ['NMR_STAR_version',     lambda x = value: returnStarString(x)],
            ['PDB_ID',               lambda x = value: returnStarString(x)],
            ['MR_file_name',         lambda x = value: returnStarString(x)],
            ['Conversion_date',      lambda x = value: returnStarString(x)]

          ]


        # Warning: key here is prefix in table tag names!
        self.sfDict[sfName].tables = [

            ['_File_characteristic', [

                ['File_characteristics_ID',  lambda x = value: returnStarInt(x)],
                ['MR_file_block_position',  lambda x = value: returnStarInt(x)],
                ['Program',    lambda x = value: returnStarString(x)],
                ['Type',    lambda x = value: returnStarString(x)],
                ['Subtype',    lambda x = value: returnStarString(x)],
                ['Format',    lambda x = value: returnStarString(x)]
              ]
            ]
          ]

      #
      # Molecular assembly definitions
      #

      sfName = 'assembly'

      if component == sfName or component == 'all':

        self.componentList.append(sfName)
        self.sfDict[sfName] = SaveFrame("",sfName)

        self.sfDict[sfName].tags = [

            ['Sf_category',   lambda x = value: x],
            ['ID',            lambda x = value: returnStarInt(x)],
            ['Name',          lambda x = value: returnStarString(x)]

          ]

        self.sfDict[sfName].tables = [

            ['_Entity_assembly', [

                ['Assembly_ID',             lambda x = value: returnStarInt(x)],
                ['ID',    lambda x = value: returnStarInt(x)],
                ['Entity_ID',    lambda x = value: returnStarInt(x)],
                ['Entity_assembly_name',    lambda x = value: returnStarString(x)],
                ['Entity_label',    lambda x = value: returnStarString(x)]
              ]
            ]
          ]


      #
      # Entity information
      #

      sfName = 'entity'

      if component == sfName or component == 'all':

        self.componentList.append(sfName)
        self.sfDict[sfName] = SaveFrame("",sfName)

        self.sfDict[sfName].tags = [

            ['Sf_category',   lambda x = value: x],
            ['ID',            lambda x = value: returnStarInt(x)],
            ['Type',          lambda x = value: x],
            ['Pol_type',      lambda x = value: x],
            ['Seq_length',    lambda x = value: returnStarInt(x)],
            ['Seq',           lambda x = value: x]
          ]

        self.sfDict[sfName].tables = [

            #
            # Describes moiety
            #

            ['_Entity_comp_index', [

                ['Entity_ID', lambda x = value: returnStarInt(x)],
                ['Num',       lambda x = value: returnStarInt(x)],
                ['Comp_ID',   lambda x = value: x]
              ]
            ],

            #
            # Describes sequence
            #

            ['_Entity_poly_seq', [

                ['Entity_ID',       lambda x = value: returnStarInt(x)],
                ['Num',             lambda x = value: returnStarInt(x)],
                ['Comp_index_num',  lambda x = value: returnStarInt(x)],
                ['Mon_ID',          lambda x = value: x]
              ]
            ]

          ]

      #
      # ChemComp information
      #

      sfName = 'chem_comp'

      if component == sfName or component == 'all':

        self.componentList.append(sfName)
        self.sfDict[sfName] = SaveFrame("",sfName)

        #
        # NOTE that IDs do NOT have to be integers!!
        #

        self.sfDict[sfName].tags = [

            ['Sf_category',   lambda x = value: x],
            ['ID',            lambda x = value: x], # is mmCif code
            ['Type',          lambda x = value: x]  # would generally be non-polymer
          ]

        self.sfDict[sfName].tables = [

            #
            # Describes moiety
            #

            ['_Chem_comp_atom', [

                ['Comp_ID',       lambda x = value: x],
                ['Atom_ID',       lambda x = value: x],
                ['Type_symbol',   lambda x = value: x]
              ]
            ]
          ]

      #
      # Distance constraints
      #

      sfName = 'distance_constraints'

      if component == sfName or component == 'all':

        self.componentList.append(sfName)
        self.sfDict[sfName] = SaveFrame("",sfName)
        self.sfDict[sfName].preTagShort = '_Dist_'

        self.sfDict[sfName].tags = [

            ['Sf_category',            lambda x = value: x],
            ['ID',                     lambda x = value: returnStarInt(x)],
            ['MR_file_block_position', lambda x = value: returnStarInt(x)],
            ['Program',                lambda x = value: x],
            ['Type',                   lambda x = value: x],
            ['Subtype',                lambda x = value: x],
            ['Format',                 lambda x = value: x]
          ]

        self.sfDict[sfName].tables = [

            ['_Dist_constraint_tree', [

                ['Constraints_ID',  lambda x = value: returnStarInt(x)],
                ['ID',              lambda x = value: returnStarInt(x)],
                ['Node_ID',         lambda x = value: returnStarInt(x)],
                ['Down_node_ID',    lambda x = value: returnStarInt(x)],
                ['Right_node_ID',   lambda x = value: returnStarInt(x)],
                ['Logic_operation', lambda x = value: returnStarString(x,strip=1)]
              ]
            ],

            ['_Dist_constraint', [

                ['Constraints_ID',                 lambda x = value: returnStarInt(x)],
                ['Dist_constraint_tree_ID',        lambda x = value: returnStarInt(x)],
                ['Tree_node_member_node_ID',       lambda x = value: returnStarInt(x)],
                ['Contribution_fractional_val',    lambda x = value: returnStarString(x)],
                ['Constraint_tree_node_member_ID', lambda x = value: returnStarInt(x)],
                ['Label_entity_assembly_ID',       lambda x = value: x],
                ['Label_entity_ID',                lambda x = value: x],
                ['Label_comp_index_ID',            lambda x = value: returnStarInt(x)],
                ['Label_comp_ID',                  lambda x = value: returnStarString(x)],
                ['Label_atom_ID',                  lambda x = value: returnStarString(x)],
                ['Auth_segment_code',              lambda x = value: returnStarString(x,strip=1)],
                ['Auth_seq_ID',                    lambda x = value: returnStarString(x)],
                ['Auth_comp_ID',                   lambda x = value: returnStarString(x,strip=1)],
                ['Auth_atom_ID',                   lambda x = value: returnStarString(x,strip=1)]
              ]
            ],

            ['_Dist_constraint_value', [

                ['Constraints_ID',            lambda x = value: returnStarInt(x)],
                ['Constraint_ID',             lambda x = value: returnStarInt(x)],
                ['Tree_node_ID',              lambda x = value: returnStarInt(x)],
                ['Source_experiment_ID',      lambda x = value: x],
                ['Spectral_peak_ID',          lambda x = value: returnStarInt(x)],
                ['Intensity_val',             lambda x = value: returnStarFloat(x)],
                ['Intensity_lower_val_err',   lambda x = value: returnStarFloat(x)],
                ['Intensity_upper_val_err',   lambda x = value: returnStarFloat(x)],
                ['Distance_val',              lambda x = value: returnStarFloat(x)],
                ['Distance_lower_bound_val',  lambda x = value: returnStarFloat(x)],
                ['Distance_upper_bound_val',  lambda x = value: returnStarFloat(x)],
                ['Weight',                    lambda x = value: returnStarFloat(x)],
                ['Spectral_peak_ppm_1',       lambda x = value: returnStarFloat(x)],
                ['Spectral_peak_ppm_2',       lambda x = value: returnStarFloat(x)]
              ]
            ],

            self.getConstraintCommentLoop(preTag = self.sfDict[sfName].preTagShort),

            self.getConstraintParseErrorLoop(preTag = self.sfDict[sfName].preTagShort),

            self.getConstraintParseFileLoop(preTag = self.sfDict[sfName].preTagShort),

            self.getConstraintConversionErrorLoop(preTag = self.sfDict[sfName].preTagShort)

          ]

      #
      # Dihedral constraints
      #

      sfName = 'torsion_angle_constraints'

      if component == sfName or component == 'all':

        self.componentList.append(sfName)
        self.sfDict[sfName] = SaveFrame("",sfName)
        self.sfDict[sfName].preTagShort = '_TA_'

        self.sfDict[sfName].tags = [

            ['Sf_category',            lambda x = value: x],
            ['ID',                     lambda x = value: returnStarInt(x)],
            ['MR_file_block_position', lambda x = value: returnStarInt(x)],
            ['Program',                lambda x = value: x],
            ['Type',                   lambda x = value: x],
            ['Subtype',                lambda x = value: x],
            ['Format',                 lambda x = value: x]
          ]

        self.sfDict[sfName].tables = [

            ['_Torsion_angle_constraint', [

                ['Constraints_ID',               lambda x = value: returnStarInt(x)],
                ['ID',                           lambda x = value: returnStarInt(x)],
                ['Torsion_angle_name',           lambda x = value: returnStarString(x)],

                ['Label_entity_assembly_ID_1',   lambda x = value: x],
                ['Label_entity_ID_1',            lambda x = value: x],
                ['Label_comp_index_ID_1',        lambda x = value: returnStarInt(x)],
                ['Label_comp_ID_1',              lambda x = value: returnStarString(x)],
                ['Label_atom_ID_1',              lambda x = value: returnStarString(x)],
                ['Label_entity_assembly_ID_2',   lambda x = value: x],
                ['Label_entity_ID_2',            lambda x = value: x],
                ['Label_comp_index_ID_2',        lambda x = value: returnStarInt(x)],
                ['Label_comp_ID_2',              lambda x = value: returnStarString(x)],
                ['Label_atom_ID_2',              lambda x = value: returnStarString(x)],
                ['Label_entity_assembly_ID_3',   lambda x = value: x],
                ['Label_entity_ID_3',            lambda x = value: x],
                ['Label_comp_index_ID_3',        lambda x = value: returnStarInt(x)],
                ['Label_comp_ID_3',              lambda x = value: returnStarString(x)],
                ['Label_atom_ID_3',              lambda x = value: returnStarString(x)],
                ['Label_entity_assembly_ID_4',   lambda x = value: x],
                ['Label_entity_ID_4',            lambda x = value: x],
                ['Label_comp_index_ID_4',        lambda x = value: returnStarInt(x)],
                ['Label_comp_ID_4',              lambda x = value: returnStarString(x)],
                ['Label_atom_ID_4',              lambda x = value: returnStarString(x)],

                ['Auth_segment_code_1',            lambda x = value: returnStarString(x,strip=1)],
                ['Auth_seq_ID_1',                  lambda x = value: returnStarString(x)],
                ['Auth_comp_ID_1',                 lambda x = value: returnStarString(x,strip=1)],
                ['Auth_atom_ID_1',                 lambda x = value: returnStarString(x,strip=1)],
                ['Auth_segment_code_2',            lambda x = value: returnStarString(x,strip=1)],
                ['Auth_seq_ID_2',                  lambda x = value: returnStarString(x)],
                ['Auth_comp_ID_2',                 lambda x = value: returnStarString(x,strip=1)],
                ['Auth_atom_ID_2',                 lambda x = value: returnStarString(x,strip=1)],
                ['Auth_segment_code_3',            lambda x = value: returnStarString(x,strip=1)],
                ['Auth_seq_ID_3',                  lambda x = value: returnStarString(x)],
                ['Auth_comp_ID_3',                 lambda x = value: returnStarString(x,strip=1)],
                ['Auth_atom_ID_3',                 lambda x = value: returnStarString(x,strip=1)],
                ['Auth_segment_code_4',            lambda x = value: returnStarString(x,strip=1)],
                ['Auth_seq_ID_4',                  lambda x = value: returnStarString(x)],
                ['Auth_comp_ID_4',                 lambda x = value: returnStarString(x,strip=1)],
                ['Auth_atom_ID_4',                 lambda x = value: returnStarString(x,strip=1)],

                ['Angle_upper_bound_val',          lambda x = value: returnStarFloat(x)],
                ['Angle_lower_bound_val',          lambda x = value: returnStarFloat(x)],
                ['Force_constant_value',           lambda x = value: returnStarFloat(x)],
                ['Potential_function_exponent',    lambda x = value: returnStarFloat(x)]
              ]
            ],

            ['_Dist_constraint', [

                ['Constraints_ID',                 lambda x = value: returnStarInt(x)],
                ['Dist_constraint_tree_ID',        lambda x = value: returnStarInt(x)],
                ['Tree_node_member_constraint_ID', lambda x = value: returnStarInt(x)],
                ['Tree_node_member_node_ID',       lambda x = value: returnStarInt(x)],
                ['Contribution_fractional_value',  lambda x = value: returnStarString(x)],
                ['Constraint_tree_node_member_ID', lambda x = value: returnStarInt(x)],
                ['Label_entity_assembly_ID',       lambda x = value: x],
                ['Label_entity_ID',                lambda x = value: x],
                ['Label_comp_index_ID',            lambda x = value: x],
                ['Label_comp_ID',                  lambda x = value: returnStarString(x)],
                ['Label_atom_ID',                  lambda x = value: returnStarString(x)],
                ['Auth_segment_code',              lambda x = value: x],
                ['Auth_seq_ID',                    lambda x = value: returnStarString(x)],
                ['Auth_comp_ID',                   lambda x = value: returnStarString(x)],
                ['Auth_atom_ID',                   lambda x = value: returnStarString(x)]
              ]
            ],


            self.getConstraintCommentLoop(preTag = self.sfDict[sfName].preTagShort),

            self.getConstraintParseErrorLoop(preTag = self.sfDict[sfName].preTagShort),

            self.getConstraintParseFileLoop(preTag = self.sfDict[sfName].preTagShort),

            self.getConstraintConversionErrorLoop(preTag = self.sfDict[sfName].preTagShort)

          ]


      #
      # RDC constraints
      #

      sfName = 'residual_dipolar_couplings'

      if component == sfName or component == 'all':

        self.componentList.append(sfName)
        self.sfDict[sfName] = SaveFrame("",sfName, prefix = '_RDC_constraints')
        self.sfDict[sfName].preTagShort = '_RDC_'

        self.sfDict[sfName].tags = [

            ['Sf_category',            lambda x = value: x],
            ['ID',                     lambda x = value: returnStarInt(x)],
            ['MR_file_block_position', lambda x = value: returnStarInt(x)],
            ['Program',                lambda x = value: x],
            ['Type',                   lambda x = value: x],
            ['Subtype',                lambda x = value: x],
            ['Format',                 lambda x = value: x]
          ]

        self.sfDict[sfName].tables = [

            ['_RDC_constraint', [

                ['Constraints_ID',               lambda x = value: returnStarInt(x)],
                ['ID',                           lambda x = value: returnStarInt(x)],
                ['Label_entity_assembly_ID_1',   lambda x = value: x],
                ['Label_entity_ID_1',            lambda x = value: x],
                ['Label_comp_index_ID_1',        lambda x = value: returnStarInt(x)],
                ['Label_comp_ID_1',              lambda x = value: returnStarString(x)],
                ['Label_atom_ID_1',              lambda x = value: returnStarString(x)],
                ['Label_entity_assembly_ID_2',   lambda x = value: x],
                ['Label_entity_ID_2',            lambda x = value: x],
                ['Label_comp_index_ID_2',        lambda x = value: returnStarInt(x)],
                ['Label_comp_ID_2',              lambda x = value: returnStarString(x)],
                ['Label_atom_ID_2',              lambda x = value: returnStarString(x)],

                ['Auth_segment_code_1',            lambda x = value: returnStarString(x,strip=1)],
                ['Auth_seq_ID_1',                  lambda x = value: returnStarString(x)],
                ['Auth_comp_ID_1',                 lambda x = value: returnStarString(x,strip=1)],
                ['Auth_atom_ID_1',                 lambda x = value: returnStarString(x,strip=1)],
                ['Auth_segment_code_2',            lambda x = value: returnStarString(x,strip=1)],
                ['Auth_seq_ID_2',                  lambda x = value: returnStarString(x)],
                ['Auth_comp_ID_2',                 lambda x = value: returnStarString(x,strip=1)],
                ['Auth_atom_ID_2',                 lambda x = value: returnStarString(x,strip=1)],

                ['RDC_val',          lambda x = value: returnStarFloat(x)],
                ['RDC_lower_bound',  lambda x = value: returnStarFloat(x)],
                ['RDC_upper_bound',  lambda x = value: returnStarFloat(x)],
                ['RDC_val_err',      lambda x = value: returnStarFloat(x)]
              ]
            ],


            self.getConstraintCommentLoop(preTag = self.sfDict[sfName].preTagShort),

            self.getConstraintParseErrorLoop(preTag = self.sfDict[sfName].preTagShort),

            self.getConstraintParseFileLoop(preTag = self.sfDict[sfName].preTagShort),

            self.getConstraintConversionErrorLoop(preTag = self.sfDict[sfName].preTagShort)

          ]

      #
      # Comments
      #

      sfName = 'MR_file_comment'

      if component == sfName or component == 'all':

        self.componentList.append(sfName)
        self.sfDict[sfName] = SaveFrame("",sfName)

        self.sfDict[sfName].tags = [

            ['Sf_category',            lambda x = value: x],
            ['ID',                     lambda x = value: returnStarInt(x)],
            ['MR_file_block_position', lambda x = value: returnStarInt(x)],
            ['Program',                lambda x = value: x],
            ['Type',                   lambda x = value: x],
            ['Subtype',                lambda x = value: x],
            ['Format',                 lambda x = value: x],
            ['Comment',                lambda x = value: returnStarString(x)]

          ]

      #
      # Coordinates
      #

      sfName = 'conformer_family_coord_set'

      if component == sfName or component == 'all':

        self.componentList.append(sfName)
        self.sfDict[sfName] = SaveFrame("",sfName)

        self.sfDict[sfName].tags = [

            ['Sf_category',            lambda x = value: x],
            ['ID',                     lambda x = value: returnStarInt(x)],
            ['Details',                lambda x = value: returnStarString(x)]

          ]

        self.sfDict[sfName].tables = [

            #
            # Describes coordinates
            #

            ['_Atom_site', [

                ['Conformer_family_coord_set_ID',   lambda x = value: returnStarInt(x)],
                ['Model_ID',                        lambda x = value: returnStarInt(x)],
                ['ID',                              lambda x = value: returnStarInt(x)],

                ['Label_entity_assembly_ID',        lambda x = value: returnStarInt(x)],
                ['Label_entity_ID',                 lambda x = value: returnStarInt(x)],
                ['Label_comp_index_ID',             lambda x = value: returnStarInt(x)],
                ['Label_comp_ID',                   lambda x = value: returnStarString(x)],
                ['Label_atom_ID',                   lambda x = value: returnStarString(x)],

                ['Auth_segment_code',               lambda x = value: returnStarString(x)],
                ['Auth_seq_ID',                     lambda x = value: returnStarString(x)],
                ['Auth_comp_ID',                    lambda x = value: returnStarString(x)],
                ['Auth_atom_ID',                    lambda x = value: returnStarString(x)],

                ['Type_symbol',                     lambda x = value: returnStarString(x)],
                ['Cartn_x',                         lambda x = value: returnStarFloat(x)],
                ['Cartn_y',                         lambda x = value: returnStarFloat(x)],
                ['Cartn_z',                         lambda x = value: returnStarFloat(x)],

                ['PDB_extracted_Bfactor_col',       lambda x = value: returnStarFloat(x)]

              ]
            ]
          ]

    #############################################
    #                                           #
    # START OF TEMPORARY VERSION 3.1 STUFF!!    #
    #                                           #
    #############################################

    elif self.version == '3.1':

      #
      # In 3.1, just read everything based on nmrStarDict...
      #

      value = 0

      try:
        from b2bTools.general.ccpn.format.nmrStar.nmrStarDict import sfDict as refSfDict
      except:
        raise ValueError("No NMR-STAR dictionary available for version 3.1!")

      for sfName in refSfDict.keys():

        #if sfName == 'assigned_chemical_shifts':
        #  debugMode = True
        #else:
        #  debugMode = False

        if component == sfName or component == 'all':
          self.componentList.append(sfName)
          self.sfDict[sfName] = SaveFrame("",sfName,prefix = "_%s" % refSfDict[sfName]['name'])

          #
          # Set tags - TODO this is very confusing - .tags is a dictionary when initialising SaveFrame. Why hacked?
          #

          self.sfDict[sfName].tags = []
          for tagName in refSfDict[sfName]['tagNames']:
            self.sfDict[sfName].tags.append([tagName,refSfDict[sfName]['tags'][tagName][1]])

          #
          # Set tables - see tags comment!
          #

          self.sfDict[sfName].tables = []

          if 'tableNames' in refSfDict[sfName]:
           for tableName in refSfDict[sfName]['tableNames']:
             self.sfDict[sfName].tables.append(['_%s' % tableName,[]])

             for tagName in refSfDict[sfName]['tables'][tableName]['tagNames']:
               self.sfDict[sfName].tables[-1][-1].append([tagName,refSfDict[sfName]['tables'][tableName]['tags'][tagName][1]])

    else:

      print ("ERROR UNRECOGNIZED VERSION '%s'" % (str(self.version)))

  def getConstraintCommentLoop(self,preTag = 'Dist_'):

    #
    # This loop occurs multiple times: defined here once
    #

    value = 0

    if self.version == '3.0':

      constraintCommentLoop = [

            preTag + 'constraint_comment_org', [

                ['Constraints_ID',       lambda x = value: returnStarInt(x)],
                ['ID',                   lambda x = value: returnStarInt(x)],
                ['Comment_begin_line',   lambda x = value: returnStarInt(x)],
                ['Comment_begin_column', lambda x = value: returnStarInt(x)],
                ['Comment_end_line',     lambda x = value: returnStarInt(x)],
                ['Comment_end_column',   lambda x = value: returnStarInt(x)],
                ['Comment',              lambda x = value: returnStarString(x)]
              ]
            ]

    elif self.version == '3.1':

      constraintCommentLoop = [

            preTag + 'constraint_comment_org', [

                ['ID',                   lambda x = value: returnStarInt(x)],
                ['Comment_text',         lambda x = value: returnStarString(x)],    # *
                ['Comment_begin_line',   lambda x = value: returnStarInt(x)],
                ['Comment_begin_column', lambda x = value: returnStarInt(x)],
                ['Comment_end_line',     lambda x = value: returnStarInt(x)],
                ['Comment_end_column',   lambda x = value: returnStarInt(x)],
              ]
            ]

    return constraintCommentLoop

  def getConstraintParseErrorLoop(self,preTag = 'Dist_'):

    #
    # This loop occurs multiple times: defined here once
    #

    value = 0

    if self.version == '3.0':

      constraintParseErrorLoop = [

            preTag + 'constraint_org_file_parse_err', [

                ['Constraints_ID',   lambda x = value: returnStarInt(x)],
                ['ID',               lambda x = value: returnStarInt(x)],
                ['Begin_line',       lambda x = value: returnStarInt(x)],
                ['Begin_column',     lambda x = value: returnStarInt(x)],
                ['End_line',         lambda x = value: returnStarInt(x)],
                ['End_column',       lambda x = value: returnStarInt(x)],
                ['Content',          lambda x = value: returnStarString(x)]
              ]
            ]

    elif self.version == '3.1':

      constraintParseErrorLoop = [

            preTag + 'constraint_parse_err', [

                ['ID',               lambda x = value: returnStarInt(x)],
                ['Content',          lambda x = value: returnStarString(x)],
                ['Begin_line',       lambda x = value: returnStarInt(x)],
                ['Begin_column',     lambda x = value: returnStarInt(x)],
                ['End_line',         lambda x = value: returnStarInt(x)],
                ['End_column',       lambda x = value: returnStarInt(x)],
              ]
            ]

    return constraintParseErrorLoop

  def getConstraintParseFileLoop(self,preTag = 'Dist_'):

    #
    # This loop occurs multiple times: defined here once - TODO 3.1
    #

    value = 0

    constraintParseFileLoop = [

          preTag + 'constraint_parse_file', [

              ['Constraints_ID',   lambda x = value: returnStarInt(x)],
              ['ID',               lambda x = value: returnStarInt(x)],
              ['Name',             lambda x = value: returnStarString(x)]
            ]
          ]

    return constraintParseFileLoop

  def getConstraintConversionErrorLoop(self,preTag = 'Dist_'):

    #
    # This loop occurs multiple times: defined here once - TODO 3.1
    #

    value = 0

    constraintConversionErrorLoop = [

          preTag + 'constraint_parse_file_conv_err', [

              ['Constraints_ID',           lambda x = value: returnStarInt(x)],
              ['ID',                       lambda x = value: returnStarInt(x)],
              ['Parse_file_ID',            lambda x = value: returnStarInt(x)],
              ['Parse_file_sf_label',      lambda x = value: returnStarString(x)],
              ['Parse_file_constraint_ID', lambda x = value: returnStarInt(x)],
              ['Conv_error_type',          lambda x = value: returnStarInt(x)],
              ['Conv_error_note',          lambda x = value: returnStarString(x)]
            ]
          ]

    return constraintConversionErrorLoop

  ###################################
  #                                 #
  # END OF VERSION 3 STUFF!!        #
  #                                 #
  ###################################

  def setupSaveFrame(self,saveFrameName,title):

    keywds = {}

    if not saveFrameName in self.sfDict:
      print ("  Warning: saveframe name %s not in reference data!" % (saveFrameName))


    else:
      keywds['prefix'] = self.getPrefix(saveFrameName)

    if not saveFrameName in self.sfs:
      self.sfs[saveFrameName] = []

    self.sfs[saveFrameName].append(SaveFrame(title,saveFrameName,**keywds))

    return self.sfs[saveFrameName][-1]

  def getPrefix(self,saveFrameName):

    return self.sfDict[saveFrameName].prefix

  def setTags(self,valueDict):

    for attr in valueDict.keys():
      value = valueDict[attr]
      setattr(self,attr,self.tags[attr][1](value))

  def readComponent(self,verbose = 0):

    #
    # WARNING: this is a silly extra level... currently upheld because
    #          fits more easily with code in importNmrChemShifts.
    #

    self.readGeneric(verbose = verbose)

    cyanaLibUsed = 0

    if self.saveFrameName in self.sfs:

      #
      # Reading all saveframes with dihedral constraint data
      #

      for saveFrame in self.sfs[self.saveFrameName]:

        dataClassFile = self.DataClassFile(self.name,self,saveFrame)
        self.files.append(dataClassFile)

        if hasattr(dataClassFile,'cyanaLibUsed') and dataClassFile.cyanaLibUsed:
          cyanaLibUsed = 1

    if not self.files:
      print ("  No %s data found in nmrStar file %s" % (self.saveFrameName,self.name))

    if cyanaLibUsed:
      print ("Using CYANA library - courtesy of Peter Guentert.")

  def readGeneric(self, verbose = 0):

    origNmrStarFile = nmrStar.File(verbosity = 2, filename = self.name)

    print ("Python nmrStar reader courtesy of Jurgen Doreleijers (BMRB) - with modifications added.")

    # Read star file in one gulp
    if origNmrStarFile.read(text = self.text):
      print ("  Error reading nmrStar file %s" % self.name)
      return

    #
    # Try to determine version...
    #

    versionHits = {'2.1.1': 0, '3.1': 0} # '3.0': 0,

    for origSaveFrame in origNmrStarFile.datanodes:

      for tagtable in origSaveFrame.tagtables:

        if tagtable.tagnames.count('_Saveframe_category'):
          versionHits['2.1.1'] += 1

        else:
          for i in range(len(tagtable.tagnames)):

            components = tagtable.tagnames[i].split(self.tagSep)

            if len(components) == 2:
              (sfName,tagName) = components

              if tagName == 'Sf_category':

                versionHits['3.1'] += 1
                break

    #
    # Reset if necessary
    #

    if not self.setVersion(versionHits):
      return

    #
    # Go through the data. Keep track of print statements in
    # conversionerror class...
    #

    convError = ConversionError()

    for origSaveFrame in origNmrStarFile.datanodes:

      saveFrameCat = None

      for tagtable in origSaveFrame.tagtables:

        try:

          tagIndex = -1

          if self.version == '2.1.1':
            tagIndex = tagtable.tagnames.index('_Saveframe_category')

          elif self.version in ('3.0','3.1'):

            for i in range(0,len(tagtable.tagnames)):

              (sfName,tagName) = tagtable.tagnames[i].split(self.tagSep)

              if tagName == 'Sf_category':

                tagIndex = i
                break

          if tagIndex != -1:
            saveFrameCat = tagtable.tagvalues[tagIndex][0]
            break

        except:
          pass


      if saveFrameCat in self.sfDict:

        saveFrame = self.setupSaveFrame(saveFrameCat,origSaveFrame.title)
        saveFrameDict = self.sfDict[saveFrameCat]

        #
        # FROM HERE ON: rerout print statements to convError object!
        #
        #
        # NOTE: if want to see error messages, best place is in setTag for SaveFrame/Table!
        #

        origStdout = sys.stdout
        sys.stdout = convError

        for tagtable in origSaveFrame.tagtables:

          #
          # Here do tags
          #

          if tagtable.free != None:

            self.setReadTags(saveFrame,saveFrameDict,tagtable,convError)

          #
          # Here do tables
          #

          else:

            self.setTableTags(saveFrame,saveFrameDict,tagtable,convError)

        sys.stdout = origStdout

    del(origNmrStarFile)

  def setVersion(self,versionHits):

    if versionHits['3.1'] > versionHits['2.1.1'] and self.version not in ('3.0','3.1'):

      print ("  Warning: setting nmrStar version to 3.1 for reading.")
      self.version = '3.1'
      self.setComponents()

    elif versionHits['2.1.1'] > versionHits['3.1'] and self.version != '2.1.1':

      print ("  Warning: setting nmrStar version to 2.1.1 for reading.")
      self.version = '2.1.1'
      self.setComponents()

    return True

  def setReadTags(self,saveFrame,saveFrameDict,tagtable,convError):

    for tagInfo in saveFrameDict.tags:

      (tagName,returnValue) = tagInfo

      matchTagName = tagName

      if self.version in ('3.0','3.1'):

        matchTagName = saveFrameDict.prefix + self.tagSep + tagName

      # TODO: Or could make set saveFrameDict in self, pass tagtable, tagName to
      # setTag...

      starValue = getNmrStarValue(tagtable,matchTagName)

      if starValue:
        value = returnValue(starValue)
      else:
        #sys.__stdout__.write(" %s, %s\n" % (matchTagName,str(tagtable.tagnames)))
        value = None

      saveFrame.setTag(tagName,value,error = convError.getString())

  def setTableTags(self,saveFrame,saveFrameDict,tagtable,convError):

    for (tableName,tags) in saveFrameDict.tables:

      tableExists = 0

      if self.version == '2.1.1' and tableName in tagtable.tagnames:
        tableExists = 1

      elif self.version in ('3.0','3.1'):

        (tempTableName,tagName) = tagtable.tagnames[0].split(self.tagSep)

        if tableName == tempTableName:
          tableExists = 1

      if tableExists:

        table = saveFrame.setupTable(tableName)
        currentTags = []

        for tagInfo in tags:

          (tagName,returnValue) = tagInfo

          matchTagName = tagName

          if self.version in ('3.0','3.1'):

            matchTagName = tableName + self.tagSep + tagName

          if matchTagName in tagtable.tagnames:

            currentTags.append([tagName,returnValue])

            table.tags[tagName] = []
            table.tagErrors[tagName] = []
            table.tagNames.append(tagName)

        for i in range(len(tagtable.tagvalues[0])):

          for tagInfo in currentTags:

            (tagName,returnValue) = tagInfo

            matchTagName = tagName

            if self.version in ('3.0','3.1'):

              matchTagName = tableName + self.tagSep + tagName

            starValue = getNmrStarValue(tagtable,matchTagName,i)
            value = returnValue(starValue)

            table.setTag(tagName,value,error = convError.getString())

  def writeGeneric(self, verbose = 0, title = None, topComment = None):

    if not title:
      title = self.name

    outputNmrStarFile = nmrStar.File(verbosity = 2, flavor = 'NMR-STAR', datanodes = [], title = title, filename = self.name)

    #
    # Jurgen credits
    #

    print ("Python nmrStar writer courtesy of Jurgen Doreleijers (BMRB) - with modifications added.")

    #
    # Write out the saveframes, use ordered list...
    #

    for saveFrameName in self.componentList:

      if not saveFrameName in self.sfs:

        continue

      for saveFrame in self.sfs[saveFrameName]:

        outputNmrStarFile.datanodes.append(nmrStar.SaveFrame(title = saveFrame.title, comment = saveFrame.comment))
        outputSaveFrame = outputNmrStarFile.datanodes[-1]

        #
        # Here do tags
        #

        outputSaveFrame.tagtables.append(nmrStar.TagTable(free = 1,tagnames = [], tagvalues = []))
        outputTagTable = outputSaveFrame.tagtables[-1]

        for tagName in saveFrame.tagNames:
          value = saveFrame.tags[tagName]

          if value != None and value != '':
            value = str(value)

            outputTagTable.tagnames.append(tagName)
            outputTagTable.tagvalues.append([value])

        #
        # Here do tables
        #

        for tableName in saveFrame.tableNames:

          table = saveFrame.tables[tableName]

          if not table.tagNames:
            if verbose:
              print ("  ERROR: nmrStar output for table %s... no tags available" % tableName)
            continue

          outputSaveFrame.tagtables.append(nmrStar.TagTable(free = None,tagnames = [], tagvalues = []))
          outputTagTable = outputSaveFrame.tagtables[-1]

          for tagName in table.tagNames:

            valFlag = False

            for value in table.tags[tagName]:

              if value is not None:
                valFlag = True
                break

            if not valFlag and verbose:
              print ('  Warning: tag %s is always None.' % tagName)

          for tagName in table.tagNames:
            outputTagTable.tagnames.append(tagName)
            outputTagTable.tagvalues.append([])
            outputTagValue = outputTagTable.tagvalues[-1]

            for value in table.tags[tagName]:

              if value == None or value == '':
                value = '.'
              else:
                value = str(value)

              outputTagValue.append(value)

    #
    # Add a top comment, if necessary
    #

    if topComment:
      outputNmrStarFile.datanodes[0].comment = topComment

    #
    # Write star file in one go
    #

    if outputNmrStarFile.write():
      print ("  Error writing nmrStar file %s" % self.name)
      return 0

    else:
      return 1

class SaveFrame:

  def __init__(self,title,name,prefix = None):

    self.tags = {}
    self.tagErrors = {}
    self.tagNames = []
    self.tables = {}
    self.tableNames = []
    self.title = title

    self.comment = ''

    self.name = name

    if prefix:

      self.prefix = prefix

    else:

      if name[0] not in string.ascii_uppercase:
        self.prefix = '_' + name.capitalize()
      else:
        self.prefix = '_' + name

  def setupTable(self,tableName):

    self.tables[tableName] = Table()
    self.tableNames.append(tableName)

    return self.tables[tableName]

  def setTag(self,tagName,value,error=None):

    if value != None and value != '':
      self.tags[tagName] = value
      self.tagErrors[tagName] = [error]

      #if error:
      #  sys.__stdout__.write("%s,%s\n" % (tagName,error))

      # Allow for overwrites...
      if tagName not in self.tagNames:
        self.tagNames.append(tagName)

  def findTableByEndText(self,tableEndText):

    tableName = None

    for tmpTableName in self.tables.keys():
      if tmpTableName.endswith(tableEndText):
        tableName = tmpTableName
        break

    return tableName

class Table:

  def __init__(self):

    self.tags = {}
    self.tagNames = []
    self.tagErrors = {}

  def setTag(self,tagName,value,error=None):

    if not tagName in self.tags:
      self.tagNames.append(tagName)
      self.tags[tagName] = []
      self.tagErrors[tagName] = []

    self.tags[tagName].append(value)
    self.tagErrors[tagName].append(error)

    #if error:
    #  sys.__stdout__.write("%s,%s\n" % (tagName,error))

class ConversionError:

  def __init__(self):

    self.string = ""

  def write(self,error):

    error = error.replace(os.linesep,'')
    error.strip()

    if error:
      self.string += error + '|'

  def getString(self):

    value = self.string[:-1]

    self.string = ""

    return value


class GenericConstraint:

  def setErrors(self,errorTags,tableTagNames,i):

    for tableTagName in tableTagNames.keys():

      if tableTagName == 'position' or not tableTagNames[tableTagName] in errorTags:
        continue

      if tableTagNames[tableTagName].find('%s') == -1:

        error = errorTags[tableTagNames[tableTagName]][i]

        if error:

          self.errors.append(error)

      else:

        for position in tableTagNames['position']:

          error = errorTags[tableTagNames[tableTagName] % position][i]

          if error:

            self.errors.append(error)

###################
# Main of program #
###################

if __name__ == "__main__":

  files = [#'../rawdata/8/1/1/info.general/bmr5106.str',
           #'../rawdata/2/2/1/info.general/bmr5307.str',
           #'../reference/nmrStar/bmr5620.str',
           #'../reference/ccpNmr/aartUtrecht/1bsh/restraints.star'#,
           #'../reference/ccpNmr/aartUtrecht/1d8v/restraints.star',
           #'../reference/ccpNmr/jurgenBmrb/1jwe/restraints.star'
           '../reference/ccpNmr/jurgenBmrb/1ao9/1ao9.str'
           ]

  for file in files:

    file = os.path.join(getTopDirectory(), file)

    nmrStarFile = NmrStarFile(file)

    nmrStarFile.readGeneric(verbose = 1)

    for key in nmrStarFile.sfs.keys():

      print (key)

      for Sf in nmrStarFile.sfs[key]:

        for tagName in Sf.tags.keys():

          print (tagName, Sf.tags[tagName])
          print (tagName, Sf.tagErrors[tagName])

        for tableName in Sf.tables.keys():

          print (tableName)

          for tagName in Sf.tables[tableName].tags.keys():

            print ("  ", tagName) #,Sf.tables[tableName].tags[tagName]

            for error in Sf.tables[tableName].tagErrors[tagName]:
              if error:
                print (error)

    #nmrStarFile.name = 'local/fulltest.star'
    #nmrStarFile.writeGeneric()



"""
======================COPYRIGHT/LICENSE START==========================

Constants.py: <write function here>

Copyright (C) 2005 Wayne Boucher, Rasmus Fogh, Tim Stevens and Wim Vranken (University of Cambridge and EBI/PDBe)

=======================================================================

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.
 
A copy of this license can be found in ../../../license/LGPL.license
 
This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.
 
You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA


======================COPYRIGHT/LICENSE END============================

for further information, please contact :

- CCPN website (http://www.ccpn.ac.uk/)
- PDBe website (http://www.ebi.ac.uk/pdbe/)

=======================================================================

If you are using this software for academic purposes, we suggest
quoting the following references:

===========================REFERENCE START=============================
R. Fogh, J. Ionides, E. Ulrich, W. Boucher, W. Vranken, J.P. Linge, M.
Habeck, W. Rieping, T.N. Bhat, J. Westbrook, K. Henrick, G. Gilliland,
H. Berman, J. Thornton, M. Nilges, J. Markley and E. Laue (2002). The
CCPN project: An interim report on a data model for the NMR community
(Progress report). Nature Struct. Biol. 9, 416-418.

Wim F. Vranken, Wayne Boucher, Tim J. Stevens, Rasmus
H. Fogh, Anne Pajon, Miguel Llinas, Eldon L. Ulrich, John L. Markley, John
Ionides and Ernest D. Laue (2005). The CCPN Data Model for NMR Spectroscopy:
Development of a Software Pipeline. Proteins 59, 687 - 696.

Rasmus H. Fogh, Wayne Boucher, Wim F. Vranken, Anne
Pajon, Tim J. Stevens, T.N. Bhat, John Westbrook, John M.C. Ionides and
Ernest D. Laue (2005). A framework for scientific data modeling and automated
software development. Bioinformatics 21, 1678-1684.

===========================REFERENCE END===============================
"""

# Tolerance for determining that two shifts are identical
shiftIdentityTolerance = 0.001
 
# conversion for nucleus NMR frequencies
# from :
# Recommendations for the presentation of NMR structures of proteins and nucleic acids
# John L. Markley, Ad Bax, Yoji Arata, C.W.Hilbers, Robert Kaptein, Brian D. Sykes, 
# Peter E. Wrightg and Kurt Wuthrich
# Journal of Biomolecular NMR, 12: 1-23, 1998
#
# The value for 19F is from the BioMagResBank web site (TFA in a sphere)
# citing Maurer, T. and Kalbitzer, H. R., 
# "Indirect Referencing of 31P and 19F NMR Spectra," 
# J. Magn. Reson. B 113, 177-178 (1996)

chemShiftRefRatios = {'1H':  1.0,
                      '2H':  0.153506088,
                      '13C': 0.251449530,
                      '15N': 0.101329118,
                      '31P': 0.404808636,
                      '19F': 0.940866982}
shiftRatio_13C_TMS = 0.25145020 # From 

#
# Reference repository name
#

refRepositoryName = 'refData'

#
# HTTP location for chemComp download
#

chemCompServer = "www.ebi.ac.uk"
chemCompWebPath = "/pdbe/docs/NMR/chemCompXml_2.0/files/"

#
# String to indicate void AtomSysName for ChemComps. This is to avoid problems with inheritance in some cases.
#

voidAtomSysName = " "

#
# Standard residue codes
#

standardResidueCcpCodes = {

    'protein':
   
      ['Ala','Arg','Asn','Asp','Cys',
       'Gln','Glu','Gly','His','Ile',
       'Leu','Lys','Met','Phe','Pro',
       'Ser','Thr','Trp','Tyr','Val'],
        
    'DNA':
    
      ['A','C','G','T'],
      
    'RNA':
    
      ['A','C','G','U']
      
                          }

code1LetterToCcpCodeDict = {

  'protein': {
     
     'A': 'Ala',
     'C': 'Cys',
     'D': 'Asp',
     'E': 'Glu',
     'F': 'Phe',
     'G': 'Gly',
     'H': 'His',
     'I': 'Ile',
     'K': 'Lys',
     'L': 'Leu',
     'M': 'Met',
     'N': 'Asn',
     'P': 'Pro',
     'Q': 'Gln',
     'R': 'Arg',
     'S': 'Ser',
     'T': 'Thr',
     'V': 'Val',
     'W': 'Trp',
     'Y': 'Tyr'
     
             },
             
  'DNA': {
     
     'A': 'A',
     'C': 'C',
     'G': 'G',
     'T': 'T',
     'I': 'I',
     'U': 'U',
     
          },
             
  'RNA': {
     
     'A': 'A',
     'C': 'C',
     'G': 'G',
     'T': 'T',
     'I': 'I',
     'U': 'U',
     
          }

    }

ccpCodeToCode1LetterDict = {

  'protein': {}, # set up below
  'DNA': code1LetterToCcpCodeDict['DNA'].copy(),
  'RNA': code1LetterToCcpCodeDict['RNA'].copy()
}

dict1 = code1LetterToCcpCodeDict['protein']
dict2 = ccpCodeToCode1LetterDict['protein']
for key in dict1.keys():
  dict2[dict1[key]] = key

#
# Information for pseudoatom corrections (from Wuthrich 1986)
#

amxPseudo =  {'HB*': (1.0, {'H': 0.6}),
              'HG*': (1.0, {'H': 1.0, 'HA': 0.6}),
              'HD*': (1.0, {'HA': 1.0}),
              'HE*': (1.0, {})}
              
aroPseudo = {'HB*': (1.0, {'H': 0.6}),
             'HD*': (2.0, {'H': 2.0, 'HA': 1.0}),
             'CG':  (2.0, {'H': 2.0, 'HA': 1.0}),
             'HE*': (2.0, {'H': 1.0, 'HA': 0.6}),
             'CZ':  (2.0, {'H': 1.0, 'HA': 0.6}),
             'HD*|HE*': (2.4, {'H': 2.4, 'HA': 2.0})}

pseudoAtomCorrectionsWuthrich = {

'Gly': {'HA*': (1.0,{})},

'Ala': {'HB*': (1.0, {'H': 0.6})},

'Ile': {'HG2*':(1.0, {'H': 1.0, 'HA': 0.6}),
        'HG1*':(1.0, {'H': 1.0, 'HA': 0.6}),
        'HD1*':(1.0, {'HA': 1.0}),
        },

'Thr': {'HG2*':(1.0, {'H': 1.0, 'HA': 0.6})},

'Val': {'HG1*':(1.0, {'H': 1.0, 'HA': 0.6}),
        'HG2*':(1.0, {'H': 1.0, 'HA': 0.6}),
        'HG*': (2.4, {'H': 1.7, 'HA': 0.6})
        },

'Leu': {'HB*': (1.0, {'H': 0.6}),
        'HD1*':(1.0, {'HA': 1.0}),
        'HD2*':(1.0, {'HA': 1.0}),
        'HD*': (2.4, {'HA': 1.7})
        },

'Lys':  amxPseudo,
'Ser':  amxPseudo,
'Asp':  amxPseudo,
'Asn':  amxPseudo,
'Cys':  amxPseudo,
'His':  amxPseudo,
'Trp':  amxPseudo,
'Glu':  amxPseudo,
'Gln':  amxPseudo,
'Met':  amxPseudo,
'Arg':  amxPseudo,
'Pro':  amxPseudo,
'Phe':  aroPseudo,
'Tyr':  aroPseudo
}


#
# Information on PDB secondary structure codes, and associated short codes
# Sometimes stored as application data on Residue level
#

secStrucInfo_kw = 'secStrucInfo'

# THIS SHOULD BECOME OBSOLETE!
ssPdbNamesDict = {             
      'Polyproline'    :    'ppr',
      'Right-handed 310':   '310',
      'Right-handed alpha': 'alf',
      'Right-handed pi':    'pih',
      'sheet'          :    'bsh',
      'turn'           :    'trn',
      None             :    'xxx'
    }


# TODO THIS SHOULD GO ELSEWHERE!!
"""
Not handled on DSSP side:
  * B = residue in isolated beta-bridge
  * S = bend 
Not handled on PDB side:
    'Polyproline'    :    '',  
"""
pdbToDsspNamesDict = {

      'Right-handed 310':   'G',
      'Right-handed alpha': 'H',
      'Right-handed pi':    'I',
      'sheet'          :    'E',
      'turn'           :    'T',
      None             :    'C',  # Is really space in DSSP.

      'Polyproline'    :    'C'  # No type available!
}


#
# Information to set up linkEnds (Wim 15/09/2004)
#

standardBackboneAtoms = {

 ('protein',):  ["H","N","CA","C","O"],
 ('DNA','RNA'): ["P","OP1","OP2","O5'","C5'","C4'","C3'","O3'","H3'"]
 
}

linkCodes = [('prev',2),('next',1)]

linkTorsions = {

  ('protein',):    [('OMEGA',['prev_2','prev_1','N','CA']),
                    ('PHI',  ['prev_1','N','CA','C']),
                    ('PSI',  ['N','CA','C','next_1'])],
             
  ('DNA','RNA'):  [('ZETA',  ['prev_2','prev_1','P',"O5'"]),
                   ('ALPHA', ['prev_1','P',"O5'","C5'"]),
                   ('EPSI',  ["C4'","C3'","O3'",'next_1'])]
  }


linkBonds = {

  ('protein',):    [('singleplanar','unknown',['prev_1','N']),
                    ('singleplanar','unknown',['C','next_1'])],
             
  ('DNA','RNA'):  [('single',None,['prev_1','P']),
                   ('single',None,["O3'",'next_1'])]
  }

linkEndDict = {

  ('protein',):    {'prev': ['N','CA'],
                    'next': ['C','CA']}, # These are 'backwards'
             
  ('DNA','RNA'):  {'prev': ['P',"O5'"],
                   'next': ["O3'","C3'"]} # These are 'backwards'
  
  }
          
linkingEnds = {

  'middle': ['prev','next'],
  'start': ['next'],
  'end': ['prev']

}



#!/usr/bin/python

"""
======================COPYRIGHT/LICENSE START==========================

dihedralConstraintsIO.py: I/O for nmrStar dihedral constraints saveframe

Copyright (C) 2005 Wim Vranken (European Bioinformatics Institute)

=======================================================================

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

A copy of this license can be found in ../../../../license/LGPL.license

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA


======================COPYRIGHT/LICENSE END============================

for further information, please contact :

- CCPN website (http://www.ccpn.ac.uk/)
- PDBe website (http://www.ebi.ac.uk/pdbe/)

- contact Wim Vranken (wim@ebi.ac.uk)
=======================================================================

If you are using this software for academic purposes, we suggest
quoting the following references:

===========================REFERENCE START=============================
R. Fogh, J. Ionides, E. Ulrich, W. Boucher, W. Vranken, J.P. Linge, M.
Habeck, W. Rieping, T.N. Bhat, J. Westbrook, K. Henrick, G. Gilliland,
H. Berman, J. Thornton, M. Nilges, J. Markley and E. Laue (2002). The
CCPN project: An interim report on a data model for the NMR community
(Progress report). Nature Struct. Biol. 9, 416-418.

Wim F. Vranken, Wayne Boucher, Tim J. Stevens, Rasmus
H. Fogh, Anne Pajon, Miguel Llinas, Eldon L. Ulrich, John L. Markley, John
Ionides and Ernest D. Laue (2005). The CCPN Data Model for NMR Spectroscopy:
Development of a Software Pipeline. Proteins 59, 687 - 696.

===========================REFERENCE END===============================
"""

import os

from b2bTools.general.ccpn.universal.Io import getTopDirectory

from b2bTools.general.ccpn.format.nmrStar.generalIO import NmrStarConstraintFile
from b2bTools.general.ccpn.format.nmrStar.generalIO import NmrStarFile
from b2bTools.general.ccpn.format.nmrStar.generalIO import GenericConstraint

#
# For DIANA torsion angle name information
#

#from b2bTools.general.ccpn.format.cyana.cyanaLibParser import CyanaLibrary

from b2bTools.general.ccpn.general.Constants import code1LetterToCcpCodeDict

from b2bTools.general.ccpn.format.general.Constants import defaultMolCode, bioPolymerCodes
from b2bTools.general.ccpn.format.general.Util import getSeqAndInsertCode

#####################
# Class definitions #
#####################

class NmrStarFile(NmrStarFile):

  def initialize(self, version = '2.1.1'):

    self.constraintFiles = []
    self.files = self.constraintFiles

    if not self.version:
      self.version = version

    self.saveFrameName = 'torsion_angle_constraints'
    self.DataClassFile = NmrStarDihedralConstraintFile

    self.components = [self.saveFrameName]
    self.setComponents()

  def read(self,verbose = 0):

    self.readComponent(verbose = verbose)

class NmrStarDihedralConstraintFile(NmrStarConstraintFile):

  def initialize(self,parent,saveFrame = None):

    self.constraints = []

    self.constraintElements = 4

    self.cyanaLibUsed = 0
    self.cyanaLib = CyanaLibrary()

    self.saveFrame = saveFrame

    self.parent = parent
    self.version = parent.version

    if self.saveFrame:
      self.parseSaveFrame()

  def parseSaveFrame(self):

    if not self.checkVersion():
      return

    #
    # Parse the constraints
    #

    tableTagNames = {}

    if self.version == '2.1.1':

      tableName = '_Constraint_ID'
      tableTagNames['constraintID'] = tableName
      tableTagNames['atomPresent'] = "_Atom_one_atom_name"
      tableTagNames['position'] = ['one','two','three','four']

      tableTagNames['name'] = '_Angle_name'
      tableTagNames['upperAngle'] = '_Angle_upper_bound_value'
      tableTagNames['lowerAngle'] = '_Angle_lower_bound_value'
      tableTagNames['energyCst'] = '_Force_constant_value'
      tableTagNames['funcExp'] = '_Potential_function_exponent'

      tableTagNames['chainCode'] = "_Atom_%s_mol_system_component_ID"
      tableTagNames['seqCode'] = "_Atom_%s_residue_seq_code"
      tableTagNames['atomName'] = "_Atom_%s_atom_name"
      tableTagNames['resLabel'] = "_Atom_%s_residue_label"

    elif self.version[0] == '3':

      tableName = '_Torsion_angle_constraint'
      tableTagNames['constraintID'] = 'ID'
      tableTagNames['position'] = ['1','2','3','4']

      tableTagNames['name'] =       'Torsion_angle_name'
      tableTagNames['upperAngle'] = 'Angle_upper_bound_val'
      tableTagNames['lowerAngle'] = 'Angle_lower_bound_val'

      if self.version == '3.0':
        tableTagNames['energyCst'] =  'Force_constant_value'
        tableTagNames['funcExp'] =    'Potential_function_exponent'

        #
        # TODO: NOTE that original input here is different from output!!!!
        # need to take that into account when mapping!!
        #

        tableTagNames['atomPresent'] = "Label_atom_ID_1"
        tableTagNames['chainCode'] = "Label_entity_assembly_ID_%s"
        tableTagNames['seqCode'] = "Label_comp_index_ID_%s"
        tableTagNames['atomName'] = "Label_atom_ID_%s"
        tableTagNames['resLabel'] = "Label_comp_ID_%s"

      elif self.version == '3.1':

        tableTagNames['atomPresent'] = "Atom_ID_1"
        tableTagNames['chainCode'] =   "Entity_assembly_ID_%s"
        tableTagNames['seqCode'] =     "Comp_index_ID_%s"
        tableTagNames['atomName'] =    "Atom_ID_%s"
        tableTagNames['resLabel'] =    "Comp_ID_%s"

      if self.saveFrame.tables.has_key(tableName) and not self.saveFrame.tables[tableName].tags[tableTagNames['seqCode'] % 1][0]:
        tableTagNames['atomPresent'] = "Auth_atom_ID_1"

        if self.version == '3.0':
          tableTagNames['chainCode'] = "Auth_segment_code_%s"
        elif self.version == '3.1':
          tableTagNames['chainCode'] = "Auth_asym_ID_%s"

        tableTagNames['seqCode'] = "Auth_seq_ID_%s"
        tableTagNames['atomName'] = "Auth_atom_ID_%s"
        tableTagNames['resLabel'] = "Auth_comp_ID_%s"

    if self.saveFrame.tables.has_key(tableName):

      constraintTableTags = self.saveFrame.tables[tableName].tags

      for i in range(0,len(constraintTableTags[tableTagNames['constraintID']])):

        dihedralConstraint = NmrStarDihedralConstraint()
        dihedralConstraint.setData(constraintTableTags,i,self.cyanaLib,tableTagNames)

        self.constraints.append(dihedralConstraint)

        dihedralConstraint.setErrors(self.saveFrame.tables[tableName].tagErrors,tableTagNames,i)

        if dihedralConstraint.name and not constraintTableTags[tableTagNames['atomPresent']][i]:

          #
          # Note that following will set item and members automatically
          #

          dihedralConstraint.setAtomMembers(constraintTableTags,i,tableTagNames)
          self.cyanaLibUsed = 1

        else:

          dihedralConstraint.nodes.append(NmrStarDihedralConstraintItem())

          for position in tableTagNames['position']:

            dihedralConstraint.nodes[-1].members.append(NmrStarDihedralConstraintMember())
            dihedralConstraint.nodes[-1].members[-1].addStarInfo(constraintTableTags,i,position,tableTagNames)

    self.parseCommentsLoop()

class NmrStarDihedralConstraint(GenericConstraint):

  def __init__(self, Id = None):

    self.Id = Id
    self.nodes = []   # THIS IS A HACK! Not really a node...
    self.errors = []

    self.name = None

  def setData(self,constraintLogicTableTags,i,cyanaLib,tableTagNames):

    self.cyanaLib = cyanaLib
    self.defaultMolCode = defaultMolCode

    self.Id = constraintLogicTableTags[tableTagNames['constraintID']][i]
    self.name = constraintLogicTableTags[tableTagNames['name']][i]
    self.upperAngle = constraintLogicTableTags[tableTagNames['upperAngle']][i]
    self.lowerAngle = constraintLogicTableTags[tableTagNames['lowerAngle']][i]

    for attrName in ('energyCst','funcExp'):
      if tableTagNames.has_key(attrName):
        setattr(self,attrName,constraintLogicTableTags[tableTagNames[attrName]][i])

  def setAtomMembers(self,constraintTableTags,i,tableTagNames):

    #
    # TODO: this is rather nasty: assuming it's DIANA if dihedral angle name given.
    #       should really be checking type... leave this for now
    #

    chainCode = constraintTableTags[tableTagNames['chainCode'] % tableTagNames['position'][0]][i]

    if not chainCode:
      chainCode = defaultMolCode

    # Can only handle real numbers here - no insert codes allowed...
    (seqCode,seqInsertCode) = getSeqAndInsertCode(constraintTableTags[tableTagNames['seqCode'] % tableTagNames['position'][0]][i])
    resLabel = constraintTableTags[tableTagNames['resLabel'] % tableTagNames['position'][0]][i]

    refAngle = self.cyanaLib.findAngle(resLabel,self.name)

    # Special casing RNA/DNA!
    if not refAngle:
      # Assuming RNA
      if len(resLabel) == 1 and resLabel[0] in code1LetterToCcpCodeDict['RNA'].keys():
        molType = 'RNA'
      # Assuming DNA
      elif len(resLabel) == 2 and resLabel[0] == 'D' and resLabel[1] in code1LetterToCcpCodeDict['DNA'].keys():
        molType = 'DNA'
      else:
        print ("    Unrecognized dihedral angle '%s' for residue label '%s' - ignoring..." % (self.name,resLabel))
        return

      for i in range(len(bioPolymerCodes[molType][0])):
        if resLabel == bioPolymerCodes[molType][0][i]:
          if molType == 'DNA':
            refCode = bioPolymerCodes[molType][1][i]
          else:
            refCode = "R%s" % bioPolymerCodes[molType][1][i]

          #print "Looking for %s for %s..." % (refCode,self.name)
          refAngle = self.cyanaLib.findAngle(refCode,self.name)
          if refAngle:
            break

    self.nodes.append(NmrStarDihedralConstraintItem())

    if refAngle:
      for atom in refAngle.atoms:

        if atom.location:
          curSeqCode = seqCode + atom.location
          curResLabel = ""

        else:
          curSeqCode = seqCode
          curResLabel = resLabel

        self.setAtomMember(self.nodes[-1],chainCode,curSeqCode,atom.name,curResLabel)

    else:
      print ("  Error: no reference angle found for name %s..." % self.name)
      for i in range(4):
        curSeqCode = seqCode
        curResLabel = resLabel
        self.setAtomMember(self.nodes[-1],chainCode,curSeqCode,'',curResLabel)

  def setAtomMember(self,node,chainCode,seqCode,atomName,resLabel):

    node.members.append(NmrStarDihedralConstraintMember(chainCode,seqCode,atomName,resLabel))

class NmrStarDihedralConstraintItem:

  def __init__(self):

    self.members = []

#
# TODO: MERGE with RDC constraints!!
#

class NmrStarDihedralConstraintMember:

  def __init__(self,chainCode = None,seqCode = None, atomName = None, resLabel = None):

    self.chainCode = chainCode
    (self.seqCode,self.seqInsertCode) = getSeqAndInsertCode(seqCode)
    self.atomName = atomName
    self.resLabel = resLabel

    self.defaultMolCode = defaultMolCode

  def addStarInfo(self,constraintTableTags,i,position,tableTagNames):

    self.chainCode = constraintTableTags[tableTagNames['chainCode'] % position][i]

    if not self.chainCode:
      self.chainCode = self.defaultMolCode

    (self.seqCode,self.seqInsertCode) = getSeqAndInsertCode(constraintTableTags[tableTagNames['seqCode'] % position][i])
    self.atomName = constraintTableTags[tableTagNames['atomName'] % position][i]
    self.resLabel = constraintTableTags[tableTagNames['resLabel'] % position][i]

###################
# Main of program #
###################

if __name__ == "__main__":

  files = [#['../reference/nmrStar/1afi.dihed.restraints','2.1.1'],
           #['../reference/ccpNmr/aartUtrecht/1byx/restraints.star','2.1.1'],
           #['../reference/ccpNmr/aartUtrecht/1bct/restraints.star','2.1.1'],
           #['../reference/ccpNmr/aartUtrecht/1d3z/restraints.star','2.1.1'],
           #['../reference/ccpNmr/aartUtrecht/1bct/restraints.star','2.1.1'],
           #['../reference/ccpNmr/aartUtrecht/1bsh/restraints.star','2.1.1'],
           #['../reference/ccpNmr/aartUtrecht/1fht/restraints.star','2.1.1'],
           #['../reference/ccpNmr/jurgenBmrb/1d8b/restraints.star','2.1.1']
           #['../reference/ccpNmr/jurgenBmrb/1n8x/1n8x.str','3.0']
           ['../reference/ccpNmr/jurgenBmrb/1ao9/1ao9.str','3.0']
          ]

  for (file,version) in files:

    file = os.path.join(getTopDirectory(), file)

    nmrStarFile = NmrStarFile(file,version = version)

    nmrStarFile.read(verbose = 1)

    for constraintFile in nmrStarFile.constraintFiles:
      #print constraintFile.comments
      for constraint in constraintFile.constraints:

        print (constraint.Id, constraint.name, constraint.lowerAngle, constraint.upperAngle)

        if constraint.errors:
          print (constraint.errors)

        for item in constraint.nodes:
          mlist = []
          for member in item.members:
            mlist.append([member.seqCode,member.seqInsertCode,member.atomName,member.resLabel])
          print ("   " + str(mlist))




"""
======================COPYRIGHT/LICENSE START==========================

util.py: Useful functions for scripts in this directory

Copyright (C) 2005-2010 Wim Vranken (European Bioinformatics Institute)

=======================================================================

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

A copy of this license can be found in ../../../../license/LGPL.license

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA


======================COPYRIGHT/LICENSE END============================

for further information, please contact :

- CCPN website (http://www.ccpn.ac.uk/)
- PDBe website (http://www.ebi.ac.uk/pdbe/)

- contact Wim Vranken (wim@ebi.ac.uk)
=======================================================================

If you are using this software for academic purposes, we suggest
quoting the following references:

===========================REFERENCE START=============================
R. Fogh, J. Ionides, E. Ulrich, W. Boucher, W. Vranken, J.P. Linge, M.
Habeck, W. Rieping, T.N. Bhat, J. Westbrook, K. Henrick, G. Gilliland,
H. Berman, J. Thornton, M. Nilges, J. Markley and E. Laue (2002). The
CCPN project: An interim report on a data model for the NMR community
(Progress report). Nature Struct. Biol. 9, 416-418.

Wim F. Vranken, Wayne Boucher, Tim J. Stevens, Rasmus
H. Fogh, Anne Pajon, Miguel Llinas, Eldon L. Ulrich, John L. Markley, John
Ionides and Ernest D. Laue (2005). The CCPN Data Model for NMR Spectroscopy:
Development of a Software Pipeline. Proteins 59, 687 - 696.

===========================REFERENCE END===============================
"""
import re

from b2bTools.general.ccpn.universal.Util import returnInt, returnFloat

anySpacePatt = re.compile("\s")
multiSpacePatt = re.compile("\s{2,}")
bracketsPatt = re.compile("[\(\)]")
yesPatt = re.compile("^[yYtT]")
noPatt = re.compile("^[nNfF]")

#############
# Functions #
#############

def removeMultiSpaces(value):

  value = str(value.strip())

  if multiSpacePatt.search(value):
    (value,numSubs) = multiSpacePatt.subn(' ',value)

  return value

def getNmrStarValue(tagtable,tagname,valueIndex = 0):

  if tagtable.tagnames.count(tagname):
    tagIndex = tagtable.tagnames.index(tagname)
    return tagtable.tagvalues[tagIndex][valueIndex]

  else:
    return None

def returnStarInt(value):

  if value in ('.','?') or value is None:
    return None
  else:
    return returnInt(value)

def returnStarFloat(value):

  if value in ('.','?') or value is None:
    return None
  else:
    return returnFloat(value)

def returnStarString(value,strip = 0, length = None):

  value = str(value)

  if strip:
    value = value.strip()

  #if value.count("\n"):
  #  value = value.replace("\n"," ")

  value = removeMultiSpaces(value)

  if length:
    origValue = value
    value = value[:length]
    if len(origValue) > length:
      print ("  Warning: shortened value '%s' to '%s' for NMR-STAR export!" % (origValue,value))

  if value in ('.','?') or not value:
    return None
  else:
    return value

def returnStarDateTime(value, length = None):
  # TODO need to do this correctly...
  return returnStarString(value,length = length)

def returnStarAtCode(value,length = None):
  # Basically accepts more or less anything, just do default.
  return returnStarString(value,length = length)

def returnStarCode(value,length = None):

  value = str(value)

  value = removeMultiSpaces(value)

  if anySpacePatt.search(value):
    (value,numSubs) = anySpacePatt.subn('_',value)

  return returnStarString(value,length = length)

def returnStarLine(value,length = None):

  value = str(value)

  #values = value.split("\n")
  #values = [ val.strip() for val in values ]
  #value = "\n".join(values)

  #if value.count("\n"):
  #  value = value.replace("\n"," ")

  value = removeMultiSpaces(value)

  value = value.strip()

  return returnStarString(value,length = length)

def returnStarName(value,length = None):

  if not value[0] == '_':
    value = '_' + value

  return returnStarCode(value,length = length)

def returnStarIdName(value,length = None):

  return returnStarCode(value,length = length)

def returnStarYesNo(value,length = None):

  if value:
    searchObj1 = yesPatt.search(str(value) )
    searchObj2 = noPatt.search(str(value) )
    if searchObj1:
      value = 'yes'
    elif searchObj2:
      value = 'no'
    else:
      value = 'yes'

  else:
    value = 'no'

  return value

def returnStarFaxPhoneEmail(value,length = None):

  value = str(value)

#  if bracketsPatt.search(value):
#    (value,numSubs) = bracketsPatt.subn('_',value)

  return returnStarString(value,length = length)

def returnStarLabel(value,length = None):

  if not value or value in ('.',):
    value =  ''

  elif not value[0] == '$':
    value = '$' + value

  return returnStarCode(value,length = length)



import numpy as np
from ..singleSeq.Predictor import MineSuite
from ..multipleSeq.msa_core import predManager
from ..multipleSeq.msa_core import BlastManager
from ..multipleSeq.msa_core import MsaManager
from ..multipleSeq.msa_core import UniRef50Manager
from ..general.plotter import Plotter
from ..singleSeq.PSPer.Predictor import PSPer as PspSuite

from ..singleSeq import constants

class SingleSeq:
    """
    Stores the sequences, the filename where they were read from, tools to be executed and the result of the
    predicitons.

    Parameters:

            fileName (str): Path to the fasta file where the sequences of interest are located.
    """

    def __init__(self, fileName):
        """
        Constructor for the SingleSeq class. Reads the fasta file that contains the sequences, calls a function that
        reads the sequences and creates the structure to store all the information required to run and store the
        predicitons.

        Parameters:

            fileName (str): Path to the fasta file where the sequences of interest are located.


        """
        self.__mineSuite__ = MineSuite()
        self.__pspSuite = PspSuite()
        self.__sequenceFileName__ = fileName
        self.__sequences__ = self.__mineSuite__.readFasta(fileName)
        self.__predictor_runners__ = {
            constants.TOOL_DYNAMINE: self.__mineSuite__.dynaMine,
            constants.TOOL_DISOMINE: self.__mineSuite__.disoMine,
            constants.TOOL_EFOLDMINE: self.__mineSuite__.eFoldMine,
            constants.TOOL_AGMATA: self.__mineSuite__.agmata,
            constants.TOOL_PSP: self.__pspSuite
        }

        self.__results__ = {}

    # This method has the responsibility of running the asked predictions in a smart way: it omits the tools that already have a result
    def predict(self, tools=[]):
        """
        This method has the responsibility of running the asked predictions in a smart way: it omits the tools that
        already have a result and runs the remaining.

        Parameters:

            tools (list[str]) : List of tools to be ran. Any combination of the following tools is accepted:
                - "dynamine"
                - "disomine"
                - "efoldmine"
                - "agmata"
                - "psp"

        Returns:

            obj: self
        """

        tools = list(set(tools))

        for tool in tools:
            dependencies = constants.DEPENDENCIES[tool]

            if (len(dependencies) > 0):
                self.predict(tools=dependencies)

            if not self.__results__.get(tool):

                predictor_function = self.__predictor_runners__[tool]
                predictor_function.allPredictions = self._all_predictions()

                if tool == constants.TOOL_EFOLDMINE:
                    predictor_function.predictSeqs(self.__sequences__, dynaMinePreds=predictor_function.allPredictions)
                elif tool == constants.TOOL_PSP:
                    seqs_dict = dict((key, value) for key, value in self.__sequences__)
                    predictor_function.predictSeqs(seqs_dict)
                else:
                    predictor_function.predictSeqs(self.__sequences__)

                self.__results__[tool] = predictor_function.allPredictions
                print("Executed {0}".format(tool.capitalize()))

        return self

    # Instead of using a list of tools as in predict method, this one uses flags to create a list of tools and call
    # the predict method:
    def semantic_predict(
            self,
            dynamics=False,
            aggregation=False,
            early_folding_propensity=False,
            disorder=False,
            phase_separating_protein=False
    ):
        """
        Instead of using a list of tools as in predict method, this one uses flags on the (high level) type of
        predictions desired to create a list of tools and call the predict method.

        Parameters:

            dynamics (bool, optional) : Whether or not dynamics predictions are to be ran
            aggregation (bool, optional) : Whether or not aggregation predictions are to be ran
            early_folding_propensity (bool, optional) : Whether or not early folding propensity predictions are to be ran
            disorder (bool, optional) : Whether or not disorder predictions are to be ran
            phase_separating_protein (bool, optional): Whether or not phase separating protein (PSP) are to be ran

        Returns:

            obj: self
        """
        tools_to_run = []

        if dynamics:
            tools_to_run.append(constants.TOOL_DYNAMINE)
        if aggregation:
            tools_to_run.append(constants.TOOL_AGMATA)
        if early_folding_propensity:
            tools_to_run.append(constants.TOOL_EFOLDMINE)
        if disorder:
            tools_to_run.append(constants.TOOL_DISOMINE)
        if phase_separating_protein:
            tools_to_run.append(constants.TOOL_PSP)

        return self.predict(tools=tools_to_run)

    def explicit_definition_predictions(self, backbone_dynamics=False,
                                        sidechain_dynamics=False,
                                        propoline_II=False,
                                        disorder_propensity=False,
                                        coil=False,
                                        beta_sheet=False,
                                        alpha_helix=False,
                                        early_folding_propensity=False,
                                        aggregation_propensity=False,
                                        protein_score=False,
                                        viterbi=False,
                                        complexity=False,
                                        arg=False,
                                        tyr=False,
                                        RRM=False):


        """

        Instead of using a list of tools as in predict method, this one uses flags on the (explicit) type of
        predictions desired to create a list of tools and call the predict method.

        Parameters:

            backbone_dynamics (bool, optional) : Whether or not backbone dynamics are to be predicted
            sidechain_dynamics (bool, optional) : Whether or not sidechain dynamics are to be predicted
            propoline_II (bool, optional) : Whether or not propoline II propensity is to be predicted
            disorder_propensity (bool, optional) : Whether or not disorder propensity is to be predicted
            coil (bool, optional) : Whether or not coil propensity is to be predicted
            beta_sheet (bool, optional) : Whether or not beta sheet propensity is to be predicted
            alpha_helix (bool, optional) : Whether or not alpha helix propensity is to be predicted
            early_folding_propensity (bool, optional) : Whether or not early folding propensity is to be predicted
            aggregation_propensity (bool, optional) : Whether or not aggregation propensity is to be predicted
            protein_score (bool, optional): Whether or not protein score from PSP is to be predicted
            viterbi (bool, optional): Whether or not Viterbi from PSP is to be predicted
            complexity (bool, optional): Whether or not complexity from PSP is to be predicted
            arg (bool, optional): Whether or not ARG from PSP is to be predicted
            tyr (bool, optional): Whether or not TYR from PSP is to be predicted
            RRM (bool, optional): Whether or not RRM from PSP is to be predicted

        Returns:

            obj: self

        """

        tools_to_run = []

        if backbone_dynamics or sidechain_dynamics or coil or beta_sheet or alpha_helix or propoline_II:
            tools_to_run.append(constants.TOOL_DYNAMINE)
        if disorder_propensity:
            tools_to_run.append(constants.TOOL_DISOMINE)
        if early_folding_propensity:
            tools_to_run.append(constants.TOOL_EFOLDMINE)
        if aggregation_propensity:
            tools_to_run.append(constants.TOOL_AGMATA)
        if protein_score or viterbi or complexity or arg or tyr or RRM:
            tools_to_run.append(constants.TOOL_PSP)

        return self.predict(tools=tools_to_run)

    def _all_predictions(self):
        all_predictions = {}

        for tool in self.__predictor_runners__:
            current_tool_predictions = self.__predictor_runners__[tool].allPredictions
            for sequence_key in current_tool_predictions:
                if not sequence_key in all_predictions:
                    all_predictions[sequence_key] = current_tool_predictions[sequence_key]
                else:
                    all_predictions[sequence_key].update(current_tool_predictions[sequence_key])

        return all_predictions

    def get_all_predictions_json(self, identifier):
        """
        Outputs all available predictions in a JSON formatted string. This still needs to be written in the desired
        output channel by the user.

        Parameters:

            identifier (str) : Identifier used as the root key of the JSON output.

        Returns:

            str : JSON string with outputs
        """
        self.__mineSuite__.allPredictions = self._all_predictions()

        return self.__mineSuite__.getAllPredictionsJson(identifier=identifier)

    def get_all_predictions(self, sequence_key=None):
        """
        Returns the values in dictionary form. It also allows to select the outputs of a single sequence from the
        original fasta file instead of all of them at once.

        Parameters:

            sequence_key (str, optional) : Sequence identifier specified as the FASTA header in the input
        file. It allow the user to select the output of a single sequence.

        Returns:

            reorganized_results (dict) : Dictionary which contains the output of the predictions.

        """
        result = self._all_predictions()

        if sequence_key is not None:
            results_to_reorganize = {sequence_key: result[sequence_key]}
        else:
            results_to_reorganize = result

        reorganized_results = self._organize_predictions_in_dictionary(results_to_reorganize)

        return reorganized_results

    def _organize_predictions_in_dictionary(self, results):
        new_result = {}
        for sequence in results:
            new_keys = {}

            for i, predictions in enumerate(results[sequence]):
                if predictions == 'seq' and 'seq' in results[sequence]:
                    new_keys['seq'] = results[sequence]['seq']
                    continue
                elif i == 0:
                    new_keys['seq'] = [position[0] for position in results[sequence][predictions]]
                if type(results[sequence][predictions]) == list or type(results[sequence][predictions]) == np.array or type(results[sequence][predictions]) == np.ndarray:
                    new_keys[predictions] = [position[1] for position in results[sequence][predictions]]
                else:
                    new_keys[predictions] = results[sequence][predictions]

            new_result[sequence] = new_keys

        return new_result

class MultipleSeq():
    """
    Class to handle all the MSA related inputs, it calculates the aligned
    predictions and plots accordingly.

    """

    def __init__(self):
        """
            Instates the predManager class required for all the methods below.

        """
        self._predManager = predManager()

    def from_aligned_file(self, path_to_msa, tools=[]):
        """
        This method has the responsibility of running the asked predictions when
        the input is an MSA file.

        Parameters:

            path_to_msa (str): Path to the file where the alignment is located.

            tools (list[str]) : List of tools to be ran. Any combination of the
            following tools is accepted:
                - "dynamine"
                - "disomine"
                - "efoldmine"
                - "agmata"

        Returns:

            obj: self
        """
        self.type = "from_msa"
        return self._predManager.run_predictors(path_to_msa, predTypes=tools)

    def from_two_msa(self, path_to_msa1, path_to_msa2, tools=[]):
        """
        This method has the responsibility of running the asked predictions when
        the input are two MSA files to be compared.

        Parameters:

            path_to_msa1 (str): Path to the 1st file where the alignment is
            located.

            path_to_msa2 (str): Path to the 2nd file where the alignment is
            located.

            tools (list[str]) : List of tools to be ran. Any combination of the
            following tools is accepted:
                - "dynamine"
                - "disomine"
                - "efoldmine"
                - "agmata"

        Returns:

            obj: self
        """
        self.type = "from_msa"
        msa = MsaManager()
        msa.t_coffe_MSA_aligner([path_to_msa1, path_to_msa2])
        return self._predManager.run_predictors(msa.out_file_name,
                                                msa_map=msa.msa_dict,
                                                predTypes=tools)

    def from_json(self, path_to_json, tools=[0], output_dir=""):
        """
        This method has the responsibility of running the asked predictions when
        the input is a json file defining sequence variants (predefined format).

        Parameters:

            path_to_json (str): Path to the json file with the sequence variants
            tools (list[str]) : List of tools to be ran. Any combination of the
            following tools is accepted:
                - "dynamine"
                - "disomine"
                - "efoldmine"
                - "agmata"
            output_dir (str): Path to the output fasta file created by reading the json file

        Returns:

            obj: self
        """
        self.type = "from_json"
        return self._predManager.run_predictors_json_input(path_to_json,
                                                           predTypes=tools)

    def from_blast(self, path_to_sequence, tools=[], blast_file_name = None, mut_option = None, mut_position=None, mut_residue=None):
        """
        This method has the responsibility of running the asked predictions when
        the input is a fasta file with a single sequence from which an alignment
        is generated with the BLAST hits.

        Parameters:

            path_to_sequence (str): Path to the fasta file where the sequence of
            interest is located.

            tools (list[str]) : List of tools to be ran. Any combination of the
            following tools is accepted:
                - "dynamine"
                - "disomine"
                - "efoldmine"
                - "agmata"

        Returns:

            obj: self
        """
        self.type = "from_msa"
        blast = BlastManager(path_to_sequence, file_name=blast_file_name, mut_option=mut_option, mut_position=mut_position, mut_residue=mut_residue)
        blast.run_qblast()
        return self._predManager.run_predictors(blast.aligned_file,
                                                mutation=None,
                                                predTypes=tools)

    def from_uniref(self, uniprotKB_id, tools=[]):
        """
        This method has the responsibility of running the asked predictions when
        the input is a UniprotKB identifier from which an alignment is generated
        with the UniRef hits (Top 25).

        Parameters:

            uniprotKB_id (str): UniprotKB identifier

            tools (list[str]) : List of tools to be ran. Any combination of the
            following tools is accepted:
                - "dynamine"
                - "disomine"
                - "efoldmine"
                - "agmata"

        Returns:

            obj: self
        """
        self.type = "from_msa"
        uniref = UniRef50Manager(uniprotKB_id, 'uniref')
        return self._predManager.run_predictors(uniref.aligned_file,
                                                predTypes=tools)

    def get_all_predictions_msa(self, sequence_key=None):
        """
        This method has the responsibility of returning the predictions mapped
        to the alignment. A sequence identifier can be used to retrieve only the
        desired predictions.

        Parameters:

            sequence_key (str): Sequence identifier

        Returns:

            obj: Aligned predictions
        """
        if sequence_key is not None:
            return self._predManager.allAlignedPredictions[sequence_key]
        else:
            return self._predManager.allAlignedPredictions

    def get_all_predictions_msa_distrib(self):
        """
        This method has the responsibility of returning the distribution of the
        predictions for all the positions in the alignment (Top-outlier, third-
        quartile, median, 1st quartile, bottom-outlier)

        Returns:

            obj: Distribution of the aligned predictions

        """
        return self._predManager.distrib

    def plot(self):
        """
        This method has the responsibility of returning the according prediction
        plots depending on the input.

        Returns:

            obj: Plots for the selected input

        """
        plotter = Plotter()
        if self.type == "from_msa":
            return plotter.plot_msa_distrib(self._predManager.jsondata_list)

        elif self.type == "from_json":
            return plotter.plot_json(self._predManager.allAlignedPredictions)
